#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include <generate_err.h>
#include <errors.h>
#include <logging.h>
#include <conv_date.h>
#include <hdr_parser.h>
#include <hdr_modify.h>

int desc_error (char *buf,char *description,int errnum)
{
	int i=0;
	
	for (i=0;(i<MAX_NUM_ERR && errnum != err[i]);i++)
		;

	if ( (i == MAX_NUM_ERR) && (errnum != err[i]) )	
	{
		strcpy (buf,"unknown");
		strcpy (description,"no description available");
		return (7);
	}
	strcpy (buf,errors[i]);
	strcpy (description,descr[i]);
	return (strlen (buf));
}

/* generate an error html page */
struct reply error_reply(const char *url, int errnum)
{
	char host[256];
	char err[MAX_ERR_LEN];
	char description[MAX_DESC_LEN];
	char header[1024];
	char *nowdate;
	struct reply ret;
	int len;
	int datalen;
	time_t tm;

	len = desc_error (err,description,errnum);

	if ( gethostname (host,sizeof(host)) == -1 )
		strcpy (host,"unknown");
	host[sizeof(host)-1]='\0';

	datalen = 449 + 2*strlen(NAME) + 2*len + 2*strlen(url) +
	          strlen(description) + strlen(VERSION) + strlen(host);

	tm = time(NULL);
	nowdate=conv_date(gmtime(&tm));

	if (nowdate==NULL)
	{
		snprintf(header,sizeof(header),"HTTP/1.1 %s\r\n"
		                               "Server: %s/%s\r\n"
		                               "Content-Length: %d\r\n"
		                               "Content-type: text/html\r\n"
		                               "Connection: close\r\n"
		                               "\r\n",
		                               err,NAME,VERSION,datalen);
	}
	else
	{
		snprintf(header,sizeof(header),"HTTP/1.1 %s\r\n"
		                               "Server: %s/%s\r\n"
		                               "Date: %s\r\n"
		                               "Expires: %s\r\n"
		                               "Content-Length: %d\r\n"
		                               "Content-type: text/html\r\n"
		                               "Connection: close\r\n"
		                               "\r\n",
		                               err,NAME,VERSION,
		                               nowdate,nowdate,datalen);
		free(nowdate);
	}
	header[sizeof(header)-1]='\0';
	
	ret.len=datalen+strlen(header);

	if ( (ret.data=malloc(ret.len+1)) == NULL )
	{
		log(LOG_ERR,"error_reply: not enough memory");
		ret.len=0;
		return ret;
	}

	strcpy (ret.data,header);
	strcat (ret.data,"<!DOCTYPE  HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">\n");
	strcat (ret.data,"<!-- this page was automatically generated by ");
	strcat (ret.data,NAME);
	strcat (ret.data,"-->\n");
	strcat (ret.data,"<HTML>\n");
	strcat (ret.data,"    <HEAD>\n");
	strcat (ret.data,"        <TITLE>Error fetching page (");
	strcat (ret.data,err);
	strcat (ret.data,")</TITLE>\n");
	strcat (ret.data,"    </HEAD>\n");
	strcat (ret.data,"    <BODY>\n");
	strcat (ret.data,"        <H1>");
	strcat (ret.data,err);
	strcat (ret.data,"</H1><HR>\n");
	strcat (ret.data,"        <P>An error has occured while trying to retrieve the requested url:\n");  
	strcat (ret.data,"        <A href=\"");
	strcat (ret.data,url);
	strcat (ret.data,"\">");
	strcat (ret.data,url);
	strcat (ret.data,"<A></P>\n");
	strcat (ret.data,"        <H3>The following error was encountered:</H3>\n");
	strcat (ret.data,"        <P>");
	strcat (ret.data,description);
	strcat (ret.data,"</P><HR>\n");
	strcat (ret.data,"        <FONT size=2><I>");
	strcat (ret.data,NAME);
	strcat (ret.data," ");
	strcat (ret.data,VERSION);
	strcat (ret.data," on ");
	strcat (ret.data,host);
	strcat (ret.data,"</I></FONT>\n");
	strcat (ret.data,"    </BODY>\n");
	strcat (ret.data,"</HTML>\n");

	return ret;
}

struct reply construct_options_reply()
{
	char header[1024];
	char *nowdate;
	struct reply ret;
	time_t tm;

	tm = time(NULL);
	nowdate=conv_date(gmtime(&tm));

	if (nowdate==NULL)
	{
		snprintf(header,sizeof(header),"HTTP/1.1 200 OK\r\n"
		                               "Server: %s/%s\r\n"
		                               "Content-Length: 0\r\n"
		                               "Allow: GET, HEAD, POST, OPTIONS, TRACE\r\n"
		                               "Connection: close\r\n"
		                               "\r\n",
		                               NAME,VERSION);
	}
	else
	{
		snprintf(header,sizeof(header),"HTTP/1.1 200 OK\r\n"
		                               "Server: %s/%s\r\n"
		                               "Date: %s\r\n"
		                               "Content-Length: 0\r\n"
		                               "Allow: GET, HEAD, POST, OPTIONS, TRACE\r\n"
		                               "Connection: close\r\n"
		                               "\r\n",
		                               NAME,VERSION,nowdate);
		free(nowdate);
	}
	header[sizeof(header)-1]='\0';

	if ( (ret.data=malloc(strlen(header)+1)) == NULL )
	{
		log(LOG_ERR,"construct_options_reply: not enough memory");
		ret.len=0;
		return ret;
	}

	strcpy (ret.data,header);
	ret.len = strlen(ret.data);

	return ret;
}

struct reply construct_trace_reply(char *request)
{
	char header[1024];
	char *nowdate;
	struct reply ret;
	int datalen;
	time_t tm;

	datalen = strlen(request);

	tm = time(NULL);
	nowdate=conv_date(gmtime(&tm));

	if (nowdate==NULL)
	{
		snprintf(header,sizeof(header),"HTTP/1.1 200 OK\r\n"
		                               "Server: %s/%s\r\n"
		                               "Content-Length: %d\r\n"
		                               "Content-type: message/http\r\n"
		                               "Connection: close\r\n"
		                               "\r\n",
		                               NAME,VERSION,datalen);
	}
	else
	{
		snprintf(header,sizeof(header),"HTTP/1.1 200 OK\r\n"
		                               "Server: %s/%s\r\n"
		                               "Date: %s\r\n"
		                               "Content-Length: %d\r\n"
		                               "Content-type: message/http\r\n"
		                               "Connection: close\r\n"
		                               "\r\n",
		                               NAME,VERSION,nowdate,datalen);
		free(nowdate);
	}
	header[sizeof(header)-1]='\0';

	ret.len=datalen+strlen(header);

	if ( (ret.data=malloc(ret.len+1)) == NULL )
	{
		log(LOG_ERR,"construct_trace_reply: not enough memory");
		ret.len=0;
		return ret;
	}

	strcpy (ret.data,header);
	strcat (ret.data,request);

	return ret;
}

struct reply immediate_answer(char **request)
{
	char hname[256];
	int r_type;
	char *host_field=NULL;
	char *via_field;
	struct reply ret;

	r_type = get_request_type(*request);

	if ( (r_type == REQ_UNKNOWN) ||
	     ( (host_field=parse_http_hdr("Host",*request)) == NULL ) )
	{
		char *url;
		url=get_requested_url(*request);
		ret=error_reply((url==NULL)?"unknown":url,400);
		free(url);
		return ret;
	}
	free(host_field);

	// Check Via fields for loops
	if ( (via_field=parse_http_hdr("Via", *request)) != NULL )
	{
		if ( gethostname (hname,sizeof(hname)) == -1 )
			strcpy (hname,"unknown-" NAME);
		hname[sizeof(hname)-1]='\0';

		if (strstr(via_field,hname) != NULL)
		{
			// There's a loop
			char *url;
			url=get_requested_url(*request);
			// Which kind of reply is suitable for a loop??
			ret=error_reply((url==NULL)?"unknown":url,400);
			free(url);
			free(via_field);
			return ret;
		}

		free(via_field);
	}

	// Max-Forwards management
	if ( (r_type == REQ_TRACE) || (r_type == REQ_OPTIONS) )
	{
		char *mf;
		if ( (mf = parse_http_hdr("Max-Forwards",*request)) != NULL )
		{
			int val;

			val=atoi(mf);

			if (val==0)
			{
				free(mf);
				switch (r_type)
				{
				case REQ_TRACE:
					return construct_trace_reply(*request);
				case REQ_OPTIONS:
					return construct_options_reply();
				}
			}
			else
			{
				char *new_hdr;

				if (!delete_header(request,"Max-Forwards"))
				{
					char *url;
					url=get_requested_url(*request);
					ret=error_reply((url==NULL)?"unknown":url,500);
					free(url);
					free(mf);
					return ret;
				}

				// If we decrement val, the new number cannot occupy more
				// space (as a string) of the previous one
				new_hdr = malloc(strlen(mf)+15);
				sprintf(new_hdr,"Max-Forwards: %d",--val);

				if (!add_header(request,new_hdr,0))
				{
					char *url;
					url=get_requested_url(*request);
					ret=error_reply((url==NULL)?"unknown":url,500);
					free(url);
					free(mf);
					free(new_hdr);
					return ret;
				}

				free(new_hdr);
			}

			free(mf);
		}
	}

	ret.data=NULL;
	return ret;
}

