#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <get_url.h>
#include <read_url.h>
#include <buffers.h>
#include <rw.h>
#include <hdr_parser.h>
#include <hdr_modify.h>
#include <logging.h>
#include <sock_utils.h>

char req_1_get[]="GET ";
char req_1_head[]="HEAD ";
char req_2[]=" HTTP/1.1\r\nHost: ";
char req_3[]="\r\n"
             "User-Agent: " NAME "/" VERSION "\r\n"
             "Accept: */*\r\n"
             "Accept-Charset: us-ascii, ISO-8859-1, ISO-8859-2, ISO-8859-4,"
             " ISO-8895-5, ISO-8895-9, ISO-8859-13, windows-1250, windows-1251,"
             " windows-1257, cp437, cp850, cp852, cp866, x-cp866-u, x-mac-ce,"
             " x-kam-cs, koi8-r, koi8-u, utf-8\r\n"
             "\r\n";
char req_3_close[]="\r\n"
                   "User-Agent: " NAME "/" VERSION "\r\n"
                   "Accept: */*\r\n"
                   "Accept-Charset: us-ascii, ISO-8859-1, ISO-8859-2,"
                   " ISO-8859-4, ISO-8895-5, ISO-8895-9, ISO-8859-13,"
                   " windows-1250, windows-1251, windows-1257, cp437,"
                   " cp850, cp852, cp866, x-cp866-u, x-mac-ce,"
                   " x-kam-cs, koi8-r, koi8-u, utf-8\r\n"
                   "Connection: close\r\n"
                   "\r\n";

int send_request_fd_hdr(int fd, char **request, int conn_close)
{
	if (!modify_header(request))
	{
		log(LOG_ERR,"send_request_fd_hdr: Cannot modify header");
		return -1;
	}

	if (conn_close)
	{
		if (!add_header(request,"Connection: close",1))
		{
			log(LOG_ERR,"send_request_fd_hdr: Not enough memory");
			return -1;
		}
	}

	// Adds our id to the "Via" header field
	if (!add_via_header(request))
	{
		log(LOG_ERR,"send_request_fd_hdr: Not enough memory");
		return -1;
	}

	/* Send request */
	if (_writen(fd,*request,strlen(*request)) < strlen(*request))
	{
		log(LOG_ERR,"send_request_fd_hdr: write error");
		return -1;
	}

	return 1;
}

// Reurns 1 or -1 if error
int send_request_fd(int fd, int type, int conn_close, char *server, char *path)
{
	char *request;
	char *r1;

	switch (type)
	{
	case REQ_GET:
		r1 = req_1_get;
		break;
	case REQ_HEAD:
		r1 = req_1_head;
		break;
	default:
		log(LOG_ERR,"send_request_fd: Unknown type (%d)", type);
		return -1;
	}

	request=malloc(strlen(r1)+strlen(req_2)+
	               strlen((conn_close)?req_3_close:req_3)+
	               strlen(server)+strlen(path)+1);

	if ( request == NULL )
	{
		log(LOG_ERR,"send_request_fd: Not enough memory");
		return -1;
	}

	strcpy(request,r1);
	strcat(request,path);
	strcat(request,req_2);
	strcat(request,server);
	strcat(request,(conn_close)?req_3_close:req_3);

	/* Send request */
	if (_writen(fd,request,strlen(request)) < strlen(request))
	{
		log(LOG_ERR,"send_request_fd: write error");
		free(request);
		return -1;
	}

	free(request);

	return 1;
}

/* fb_len holds the length of the yet-readen buffer (first_buf).
 * When this function returns fb_len is the actual size of the returned doc. */
char *read_document_fd(int fd, char **headers, char *first_buf, int *fb_len,
                       int *reply_fd)
{
	char *ret;
	char *len;

	if ( (len=parse_http_hdr("Content-Length",*headers)) != NULL )
	{
		/* WARNING: atoi w/o checking here is DANGEROUS (possible DoS) */
		if ( (ret=read_sized(fd,first_buf,*fb_len,atoi(len),reply_fd)) == NULL )
		{
			free(len);
			*fb_len = 0;
			return NULL;
		}
		*fb_len=atoi(len);
	}
	else if ( ( (len=parse_http_hdr("Transfer-Encoding",*headers)) != NULL ) &&
	          ( strstrcase(len,"chunked") != NULL ) )
	{
		if ( (ret=read_chunked(fd,first_buf,*fb_len,headers,reply_fd)) == NULL )
		{
			free(len);
			*fb_len = 0;
			return NULL;
		}
		free(len);
		if ( (len=parse_http_hdr("Content-Length",*headers)) == NULL )
		{
			free(ret);
			*fb_len = 0;
			return NULL;
		}
		*fb_len=atoi(len);
	}
	else
	{
		/* Cannot estabilish when the document is ended, so go on
		 * reading expecting the end of the connection */
		log(LOG_WARNING,"read_document_fd: malformed reply (Server is 1.0?)");
		// fb_len is set as the length of the readen document inside
		// this function
		if ( (ret=read_connclose(fd,first_buf,fb_len,reply_fd)) == NULL )
		{
			*fb_len = 0;
			free(len);
			return NULL;
		}
	}

	free(len);
	return ret;
}

/* Retrieves headers of a document over an existing connection */
char *get_headers_fd(int fd, char *server, char *path, int conn_close,
                     int *reply_fd)
{
	char *ptr;

	if (send_request_fd(fd,REQ_HEAD,conn_close,server,path) < 0)
	{
		log(LOG_ERR,"get_headers_fd: cannot send request");
		return NULL;
	}

	/* Read header */
	if ( (ptr=read_headers(fd,NULL,0,reply_fd)) == NULL )
	{
		log(LOG_ERR,"get_headers_fd: error reading headers");
		return NULL;
	}

	return ptr;
}

/* Retrieves a document over an existing connection */
struct document get_document_fd(int fd, char *server, char *path,
                                int conn_close, char **request, int *reply_fd)
{
	char *ptr=NULL;
	int read_len;
	struct document ret;

	ret.doc = ret.hdr = NULL;
	ret.doclen = -1;

	if (request==NULL)
	{
		// Prefetching: use standard request headers
		if (send_request_fd(fd,REQ_GET,conn_close,server,path) < 0)
		{
			log(LOG_ERR,"get_document_fd: cannot send request");
			return ret;
		}
	}
	else
	{
		if (send_request_fd_hdr(fd,request,conn_close) < 0)
		{
			log(LOG_ERR,"get_document_fd: cannot send request");
			return ret;
		}
	}

	/* Read header */
	if ( (ret.hdr=read_headers(fd,&ptr,&read_len,reply_fd)) == NULL )
	{
		log(LOG_ERR,"get_document_fd: error reading headers");
		return ret;
	}

	if (there_is_data((request==NULL)?NULL:*request,ret.hdr))
	{
		ret.doc=read_document_fd(fd, &ret.hdr, ptr, &read_len, reply_fd);

		ret.doclen=read_len;

		if (ret.doc == NULL)
		{
			free(ret.hdr);
			ret.hdr = NULL;
		}
		free(ptr);
	}
	else
	{
		free(ptr);
		ret.doc=NULL;
		ret.doclen=0;
	}

	return ret;
}

struct document get_document(char *server, char *resolved_srv, int port,
                             char *path, char **request, int reply_fd,
                             int connect_timeout)
{
	struct document ret;
	int err;
	int fd;

	/* Open connection */
	err=tcp_open(&fd,( (resolved_srv==NULL)?server:resolved_srv ),port,
	             connect_timeout);
	if (err != 0)
	{
		log(LOG_WARNING,"get_document: Cannot connect to %s",
		                ((resolved_srv==NULL)?server:resolved_srv));
		ret.doc=ret.hdr=NULL;
		ret.doclen=0;
		return ret;
	}

	/* Always close connection for now */
	ret=get_document_fd(fd,server,path,1,request,&reply_fd);
	close(fd);

	return ret;
}

char *get_headers(char *server, char *resolved_srv, int port, char *path,
                  int reply_fd)
{
	char *ret;
	int err;
	int fd;

	/* Open connection */
	err=tcp_open(&fd,( (resolved_srv==NULL)?server:resolved_srv ),port,10);
	if (err != 0)
	{
		log(LOG_WARNING,"get_document: Cannot connect to %s",server);
		return NULL;
	}

	/* Always close connection for now */
	ret=get_headers_fd(fd,server,path,1, &reply_fd);
	close(fd);

	return ret;
}

