#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <struct.h>
#include <hash.h>
#include <init_shared.h>
#include <alloc.h>
#include <logging.h>
#include <rw.h>
#include <sem.h>
#include <expire.h>

/* The argument x is an address of memory:
 * if x is a relative address of shared memory then
 * REAL_ADDR convert x into an absolute address of shared memory.
 * If x is not an address of shared memory
 * REAL_ADDR return x.
 * We must insert relative address in pointers stored in shared memory.
 */
#define REAL_ADDR(x) ( ( ((x) > ((typeof (x))0)) &&                       \
                         ((x) < ((typeof(x))(shmem_size+2))) ) ?          \
                           ( (typeof (x))(shmem_start + (int)(x) - 1) ) : \
					       (x) )

//Utility function.
char* create_path_local(char *doc, char *host, char *port);
char* create_path(char *doc, char *host, char *port);
void* allocate(int size, int need_P);

//Cache management function.
int hash_func(char *host_name);
struct hash_host* find_host(char *host);
struct hash_doc* find_doc(struct hash_host *host, char *path, int port);
struct hash_doc* search_doc(char *host, char *path, int port);
struct hash_host* insert_host(char *host);
int remove_host(char *host);
int remove_file(char *host, char *doc);
int replace_doc(struct document *doc, struct urlinfo_t *url);
int save_doc(char *name, char *buf, int size);
struct document* load_doc(struct hash_doc *doc);

//Struct search_dl management function.
void delete_search_dl(struct search_dl *tmp);
struct search_dl* find_search_dl(char *host, char *doc, int port);
int insert_search_dl(struct urlinfo_t *url);

struct hash_host **hash; //hash table
struct search_dl **queue; //hold the process blocked waiting one document
int hash_mutex; //sem for hash mutex
int queue_mutex; //sem fo search_dl mutex

//Calculate hash function.
//Return the key value.
int hash_func(char *host_name)
{
	int somma_hn = 0;
	char *s;
	for ( s = host_name; *REAL_ADDR(s) != 0; s++)
		somma_hn += *REAL_ADDR(s);
	return ( somma_hn % HASHTABLESIZE );
}

//Allocate the hash table and create the general directory on filesystem.
//Return 1 on success, 0 otherwise.
int init_hash(void)
{
	int i;
	int size;
	size = sizeof(struct hash_host *) * HASHTABLESIZE;
	if ( (hash = allocate(size, 1)) == NULL ) //allocate table hash
	{
		log(LOG_ERR, "init_hash: Cannot allocate shared memory");
		return 0;
	}
	for ( i = 0; i < HASHTABLESIZE; i++ )
		REAL_ADDR(hash)[i] = NULL; //initialize host list
	if ( access(CACHEDIR, F_OK)  < 0 )
		if ( mkdir(CACHEDIR, PERMD) < 0 ) //create main dircetory
		{
			log(LOG_ERR, "init_hash: Cannot create main directory");
			return 0;
		}
	if ( (hash_mutex = create_sem(1)) < 0 ) //create mutex semaphore
	{
		log(LOG_ERR, "init_hash: Cannot create mutex semaphore");
		return 0;
	}
	if ( (queue_mutex = create_sem(1)) < 0 ) //create mutex semaphore
	{
		log(LOG_ERR, "init_hash: Cannot create mutex semaphore");
		return 0;
	}
	//allocate search_dl start ptr
	if ( (queue = allocate(sizeof(struct search_dl *), 1)) == NULL )
	{
		log(LOG_ERR, "init_hash: Cannot allocate queue pointer");
		return 0;
	}
	*REAL_ADDR(queue) = NULL;
	return 1;
}

//Convert port to a string. Don't use shared memory.
//Return pointer to string on success, NULL on failure.
char* port_to_string_local(int port)
{
	char temp[10];
	char *ret;
	snprintf(temp,sizeof(temp),"%d",port);
	temp[sizeof(temp)-1] = '\0';
	if ( (ret = malloc(strlen(temp)+1)) == NULL )
	{
		log(LOG_ERR,"port_to_string_local: Cannot allocate memory");
		return NULL;
	}
	strcpy(ret, temp);
	return ret;
}

//Convert port to a string. Use shared memory.
//Return pointer to string on success, NULL on failure.
char* port_to_string(int port)
{
	char temp[10];
	char *ret;
	snprintf(temp,sizeof(temp),"%d",port);
	temp[sizeof(temp)-1] = '\0';
	if ( (ret = allocate(strlen(temp)+1, 0)) == NULL )
	{
		log(LOG_ERR,"port_to_string: Cannot allocate memory");
		return NULL;
	}
	strcpy(REAL_ADDR(ret),temp);
	return ret;
}

//Create path on filesystem. Don't use shared memory.
//Return the path on success, NULL on failure,
char* create_path_local(char *doc, char *host, char *port)
{
	int i, size, sizec;
	char *buf;
	sizec = strlen(CACHEDIR);
	size = sizec + strlen(REAL_ADDR(host)) + strlen(REAL_ADDR(doc));
	if ( REAL_ADDR(host)[strlen(REAL_ADDR(host))-1] != '/' )
		size += 1;
	if (port != NULL)
		size += strlen(REAL_ADDR(port))+1;
	if ( (buf = malloc(size+1)) == NULL )
	{
		log(LOG_ERR, "create_path_local: Cannot allocate memory");
		return NULL;
	}
	strcpy(buf, CACHEDIR); //add directory cache name to path
	strcat(buf, REAL_ADDR(host)); //add host name to path
	if ( buf[strlen(buf)-1] != '/' )
		strcat(buf, "/");
	i = strlen(buf);
	while ( ( *REAL_ADDR(doc) != 0 ) && ( *REAL_ADDR(doc) != '#' ) )
	{
		buf[i] = *REAL_ADDR(doc); //add doc path
		if ( buf[i] == '/' )
			buf[i] = '_';
		i++;
		doc++;
	}
	if (port != NULL)
	{
		buf[i++] = '_';
		buf[i] = '\0';
		strcat(buf,REAL_ADDR(port));
	}
	else
		buf[i] = '\0';
	return buf;
}

//Create path on filesystem; use shared memory.
//Return the path on success, NULL on failure,
char* create_path(char *doc, char *host, char *port)
{
	int i, size, sizec;
	char *buf;
	sizec = strlen(CACHEDIR);
	size = sizec + strlen(REAL_ADDR(host)) + strlen(REAL_ADDR(doc));
	if ( REAL_ADDR(host)[strlen(REAL_ADDR(host))-1] != '/' )
		size += 1;
	if (port != NULL)
		size += strlen(REAL_ADDR(port))+1;
	if ( (buf = allocate(size+1,0)) == NULL )
	{
		log(LOG_ERR, "create_path: Cannot allocate memory");
		return NULL;
	}
	strcpy(REAL_ADDR(buf), CACHEDIR); //add directory cache name to path
	strcat(REAL_ADDR(buf), REAL_ADDR(host)); //add host name to path
	if ( REAL_ADDR(buf)[strlen(REAL_ADDR(buf))-1] != '/' )
		strcat(REAL_ADDR(buf), "/");
	i = strlen(REAL_ADDR(buf));
	while ( ( *REAL_ADDR(doc) != 0 ) && ( *REAL_ADDR(doc) != '#' ) )
	{
		REAL_ADDR(buf)[i] = *REAL_ADDR(doc); //add doc path
		if ( REAL_ADDR(buf)[i] == '/' )
			REAL_ADDR(buf)[i] = '_';
		i++;
		doc++;
	}
	if (port != NULL)
	{
		REAL_ADDR(buf)[i++] = '_';
		REAL_ADDR(buf)[i] = '\0';
		strcat(REAL_ADDR(buf),REAL_ADDR(port));
	}
	else
		REAL_ADDR(buf)[i] = 0;
	return buf;
}

//Try to allocate memory until the sh_malloc doesn't return NULL
//Return pointer to memiry on success, NULL on failure.
void* allocate(int size, int need_P)
{
	char *path;
	void *ret;
	struct hash_host *h;
	struct hash_doc *d;
	struct search_dl *sdl;
	int host, doc, waiting = 0;
	ret = sh_malloc(size);
	while ( ret == NULL ) //not enough memory
	{
		host = random() % HASHTABLESIZE;
		doc = random() % MAXDOC; //doc to remove for free space
		if (need_P)
			if ( sem_P(hash_mutex) < 0 )
			{
				log(LOG_ERR, "allocate: P failed");
				return NULL;
			}
		h = REAL_ADDR(hash)[host];
		if ( h != NULL ) //host in cache
		{
			d = REAL_ADDR(h)->list[doc];
			if ( d != NULL ) //doc in cache
			{
				sdl = *REAL_ADDR(queue);
				while ( (sdl != NULL) && (!waiting) )
				{
					if ( (path = create_path_local(REAL_ADDR(sdl)->doc,
					      REAL_ADDR(sdl)->host,
					      port_to_string_local(REAL_ADDR(sdl)->port))) == NULL )
					{
						if (need_P)
							if ( sem_V(hash_mutex) < 0 )
								log(LOG_ERR, "allocate: V failed");
						return NULL;
					}
					if ( strcmp(path,REAL_ADDR(REAL_ADDR(d)->path)) == 0 )
						waiting = 1;
					else
						sdl = REAL_ADDR(sdl)->next;
					free(path);
				}
				if ( ! waiting ) //there aren't process blocked on this document
				{
					if ( remove_file(REAL_ADDR(h)->host_name,REAL_ADDR(d)->path)
					     == 0 ) //free space
					{
						if (need_P)
							if ( sem_V(hash_mutex) < 0 )
								log(LOG_ERR, "allocate: V failed");
						return NULL;
					}
					ret = sh_malloc(size); //retry to allocate memory
				}
			}
		}
		if (need_P)
			if ( sem_V(hash_mutex) < 0 )
			{
				log(LOG_ERR, "allocate: V failed");
				return NULL;
			}
	}
	return ret;
}

//Get the header associated to url.
//Return NULL on error, the pointer to header otherwise.
char* get_cache_header(struct urlinfo_t *url)
{
	struct hash_doc *doc;
	if ( sem_P(hash_mutex) < 0 )
	{
		log(LOG_ERR, "get_cache_header: P failed");
		return NULL;
	}
	doc = search_doc(url->resolved_host, url->path, url->port);
	if ( sem_V(hash_mutex) < 0 )
	{
		log(LOG_ERR, "get_cache_header: V failed");
		return NULL;
	}
	if ( doc == NULL ) //doc not in cache
		return NULL;
	else //doc in cache
		return REAL_ADDR(doc)->header;
}

//Search the host element on the cache.
//Return NULL if it isn't on the cache, the pointer to element otherwise.
struct hash_host* find_host(char *host)
{
	int pos;
	struct hash_host *tmp;
	pos = hash_func(host);
	tmp = REAL_ADDR(hash)[pos]; //position on table hash
	while ( (tmp != NULL) && (REAL_ADDR(tmp)->host_name != NULL) &&
	      (strcmp(REAL_ADDR(REAL_ADDR(tmp)->host_name), REAL_ADDR(host)) != 0) )
	{
		tmp = REAL_ADDR(tmp)->next;
	}
	return tmp;
}

//Search the document on the cache.
//Return pointer to doc if is on the cache, NULL otherwise.
struct hash_doc* find_doc(struct hash_host *host, char *path, int port)
{
	int i = 0, b = 1;
	char *tmp;
	tmp = create_path(path, REAL_ADDR(host)->host_name,
			port_to_string(port));//doc name on filesystem
	while ( (i < MAXDOC) && b )
	{
		if ( (REAL_ADDR(host)->list[i] != NULL) &&
		     (REAL_ADDR(REAL_ADDR(host)->list[i])->path != NULL) )
			if ( !strcmp(REAL_ADDR(REAL_ADDR(REAL_ADDR(host)->list[i])->path),
			             REAL_ADDR(tmp)) )
				b = 0;
		i++;
	}
	if ( !b )
		i--;
	sh_free(tmp);
	if ( i == MAXDOC ) //doc not in cache
		return NULL;
	return REAL_ADDR(host)->list[i];
}

//Search document on the cache.
//Return NULL if it doesn't find host or document, pointer to document otherwise.
struct hash_doc* search_doc(char *host, char *path, int port)
{
	struct hash_host *s_host;
	if ( (s_host = find_host(host)) == NULL ) //host not in cache
		return NULL;
	return find_doc(s_host, path, port);
}

//Insert (to head) host in cache on memory.
//Return pointer to host on success, NULL otherwise.
struct hash_host* insert_host(char *host)
{
	char *dir;
	struct hash_host *ins;
	int pos, j, size;
	pos = hash_func(host);
	if ( (ins = allocate(sizeof(struct hash_host),0)) == NULL ) //allocate host
	{
		log(LOG_ERR, "insert_host: Cannot allocate memory");
		return NULL;
	}
	if ( (REAL_ADDR(ins)->host_name = allocate(strlen(REAL_ADDR(host))+1,0)) == NULL )
	{
		log(LOG_ERR, "insert_host: Cannot allocate memory");
		return NULL;
	}
	strcpy(REAL_ADDR(REAL_ADDR(ins)->host_name),REAL_ADDR(host));//set host name
	for ( j = 0; j < MAXDOC; j++ )
		REAL_ADDR(ins)->list[j] = NULL; //initialize document list
	REAL_ADDR(ins)->next = REAL_ADDR(hash)[pos]; //set next element in this list
	REAL_ADDR(hash)[pos] = ins; //insert element into hash table
	size = strlen(CACHEDIR) + strlen(REAL_ADDR(host)) + 1;
	if ( (dir = malloc(size)) == NULL )
	{
		log(LOG_ERR, "insert_host: Cannot allocate memory");
		return NULL;
	}
	strcpy(dir, CACHEDIR);
	strcat(dir, REAL_ADDR(host)); //host directory name
	if ( access(dir, F_OK) < 0 )
		if ( mkdir(dir, PERMD) < 0 ) //create host directory
		{
			log(LOG_ERR, "insert_host: Cannot create host directory");
			free(dir);
			return NULL;
		}
	free(dir);
	return ins;
}

//Insert file on the filesystem cache.
//Retun 1 on success, 0 or -1 otherwise.
int insert_file(struct urlinfo_t *url, struct document *data)
{
	struct hash_host *ins;
	struct hash_doc *doc;
	struct search_dl *sdl;
	struct urlinfo_t rem;
	int i = 0;

	if ( sem_P(hash_mutex) < 0 )
	{
		log(LOG_ERR, "insert_file: P failed");
		return 0;
	}
	ins = find_host(url->resolved_host);

	if ( ins == NULL ) //host not in cache, we insert it
		if ( (ins = insert_host(url->resolved_host)) == NULL )
		{
			if ( sem_V(hash_mutex) < 0 )
			{
				log(LOG_ERR, "insert_file: V failed");
				return 0;
			}
			return 0;
		}
	if ( (doc = search_doc(url->resolved_host, url->path, url->port)) == NULL )
	{
		if ( (doc = allocate(sizeof(struct hash_doc),0)) == NULL ) //allocate document
		{
			log(LOG_ERR, "insert_file: Cannot allocate memory");
			if ( sem_V(hash_mutex) < 0 )
			{
				log(LOG_ERR, "insert_file: V failed");
				return 0;
			}
			return 0;
		}
		if ( (REAL_ADDR(doc)->header = allocate(strlen(data->hdr)+1,0)) == NULL )
		{
			log(LOG_ERR, "insert_file: Cannot allocate memory");
			if ( sem_V(hash_mutex) < 0 )
			{
				log(LOG_ERR, "insert_file: V failed");
				return 0;
			}
			return 0;
		}
		if ((REAL_ADDR(doc)->url_path=allocate(strlen(url->path)+1,0)) == NULL )
		{
			log(LOG_ERR, "insert_file: Cannot allocate memory");
			if ( sem_V(hash_mutex) < 0 )
			{
				log(LOG_ERR, "insert_file: V failed");
				return 0;
			}
			return 0;
		}
		strcpy(REAL_ADDR(REAL_ADDR(doc)->header), data->hdr); //set header
		strcpy(REAL_ADDR(REAL_ADDR(doc)->url_path), url->path); //set url->path
		REAL_ADDR(doc)->size = url->size; //set size
		REAL_ADDR(doc)->port = url->port; //set port
		if ( (REAL_ADDR(doc)->path = create_path(url->path, url->resolved_host,
		         port_to_string(url->port))) == NULL ) //doc path on filesystem
		{
			if ( sem_V(hash_mutex) < 0 )
			{
				log(LOG_ERR, "insert_file: V failed");
				return 0;
			}
			return 0;
		}
		while ( (i < MAXDOC) && (REAL_ADDR(ins)->list[i] != NULL) )
			i++;
		rem.port = -1;
		if ( i == MAXDOC ) //host list full
		{
			i=0;
			while ( (i<MAXDOC) && (!is_expired(REAL_ADDR(REAL_ADDR(
			                             REAL_ADDR(ins)->list[i])->header))) )
			{
				i++;
			}
			if ( i == MAXDOC )
				i = random() % MAXDOC;
			strcpy(rem.host, REAL_ADDR(REAL_ADDR(ins)->host_name));
			strcpy(rem.path,REAL_ADDR(REAL_ADDR(REAL_ADDR(ins)->list[i]->url_path)));
			rem.port = REAL_ADDR(REAL_ADDR(ins)->list[i])->port;
			if ( remove_file(REAL_ADDR(ins)->host_name,
			   	         REAL_ADDR(REAL_ADDR(ins)->list[i])->path) == 0)
			{
				if ( sem_V(hash_mutex) < 0 )
				{
					log(LOG_ERR, "insert_file: V failed");
					return 0;
				}
				return 0;
			}
		}
		REAL_ADDR(ins)->list[i] = doc; //insert doc in the host list
		if ( save_doc(REAL_ADDR(REAL_ADDR(doc)->path), data->doc, data->doclen) == 0 ) //write on filesystem
		{
			if ( sem_V(hash_mutex) < 0 )
			{
				log(LOG_ERR, "insert_file: V failed");
				if ( rem.port >= 0 )
					clear_one(&rem);
				return 0;
			}
			if ( rem.port >= 0 )
				clear_one(&rem);
			return 0;
		}
		if ( sem_V(hash_mutex) < 0 )
		{
			log(LOG_ERR, "insert_file: V failed");
			if ( rem.port >= 0 )
				clear_one(&rem);
			return 0;
		}
		if ( rem.port >= 0 )
			if ( !clear_one(&rem) )
				return 0;
		if ( sem_P(queue_mutex) < 0 )
		{
			log(LOG_ERR, "insert_file: P failed");
			return 0;
		}
		if ( (sdl = find_search_dl(url->resolved_host, url->path, url->port)) != NULL )
		{
			if ( REAL_ADDR(sdl)->cont ) //there are blocked process
			{
				if ( sem_V(REAL_ADDR(sdl)->sem) < 0 )
				{
					log(LOG_ERR, "insert_file: V failed");
					if ( sem_V(queue_mutex) < 0 )
						log(LOG_ERR, "insert_file: V failed");
					return 0;
				}
			}
			else //there aren't blocked process
				delete_search_dl(sdl);
		}
		if ( sem_V(queue_mutex) < 0 )
		{
			log(LOG_ERR, "insert_file: V failed");
			return -1;
		}
	}
	else
	{
		if (!replace_doc(data,url))
		{
			if ( sem_V(hash_mutex) < 0 )
			{
				log(LOG_ERR, "insert_file: V failed");
				return -1;
			}
			return -1;
		}

		if ( sem_V(hash_mutex) < 0 )
		{
			log(LOG_ERR, "insert_file: V failed");
			return -1;
		}
	}
	return 1;
}

//Save document on the filesystem.
//Return 1 on success, 0 on failure.
//name and buf must be real address, not relative
int save_doc(char *name, char *buf, int size)
{
	int fd;
	if ( creat(name, PERMF) < 0 ) //create the file
	{
		log(LOG_ERR, "save_doc: Cannot creat file");
		return 0;
	}
	if ( (fd = open(name, O_RDWR)) < 0 ) //open the file
	{
		log(LOG_ERR, "save_doc: Cannot open file");
		return 0;
	}
	if ( _writen(fd, buf, size) < size ) //save to filesystem
	{
		log(LOG_ERR, "save_doc: Cannot write file");
		return 0;
	}
	close(fd); //close document
	return 1;
}

//Load documento from filesystem to memory.
//Return the document on success, NULL otherwise.
struct document* load_doc(struct hash_doc *doc)
{
	struct document *ret;
	int fd, e;
	if ( (fd = open(REAL_ADDR(REAL_ADDR(doc)->path), O_RDWR)) < 0 ) //open the document
	{
		log(LOG_ERR, "load_doc: Cannot open file");
		return NULL;
	}
	if ( (ret = malloc(sizeof(struct document))) == NULL ) //allocate the memory
	{
		log(LOG_ERR,"load_doc: Cannot allocate memory");
		return NULL;
	}
	if ( (ret->doc = malloc(REAL_ADDR(doc)->size)) == NULL )
	{
		log(LOG_ERR, "load_doc: Cannot allocate shared memory");
		free(ret);
		return NULL;
	}
	if ( (ret->hdr = malloc(strlen(REAL_ADDR(REAL_ADDR(doc)->header))+1)) == NULL )
	{
		log(LOG_ERR, "load_doc: Cannot allocate shared memory");
		free(ret->doc);
		free(ret);
		return NULL;
	}
	strcpy(ret->hdr, REAL_ADDR(REAL_ADDR(doc)->header)); //set header on the structure to return
	if ( _readn(fd, ret->doc, REAL_ADDR(doc)->size) < REAL_ADDR(doc)->size ) //load document in memory
	{
		log(LOG_ERR, "load_doc: Cannot read file");
		free(ret->doc);
		free(ret->hdr);
		free(ret);
		return NULL;
	}
	if ( (e = is_expired(REAL_ADDR(REAL_ADDR(doc)->header))) == 0 ) //doc not expired
		ret->valid = 1;
	else if ( e == -1 ) //error
	{
		log(LOG_ERR, "load_doc: Error on expiration");
		return NULL;
	}
	else //doc expired or to validate
		ret->valid = 0;
	ret->doclen = REAL_ADDR(doc)->size;
	close(fd); //close the document
	return ret;
}

//Give the document to proxy.
//Return pointer to document if is on the cache (also if is expired).
//Return -1 if the doc is not in cache, NULL on error.
struct document* get_doc(struct urlinfo_t *url)
{
	struct hash_doc *doc;
	struct document *ret;
	struct search_dl *tmp;
	if ( sem_P(hash_mutex) < 0 )
	{
		log(LOG_ERR, "get_doc: P failed");
		return NULL;
	}
	if ( (doc = search_doc(url->resolved_host, url->path,
	                       url->port)) == NULL ) //doc not in cache
	{
		if ( sem_V(hash_mutex) < 0 )
		{
			log(LOG_ERR, "get_doc: V failed");
			return NULL;
		}
		if ( sem_P(queue_mutex) < 0 )
		{
			log(LOG_ERR, "get_doc: P failed");
			return NULL;
		}
		if ( (tmp = find_search_dl(url->resolved_host, url->path, url->port)) == NULL ) //struct not present
		{
			if ( insert_search_dl(url) == 0 ) //creat and insert element on struct
			{
				if ( sem_V(queue_mutex) < 0 )
				{
					log(LOG_ERR, "get_doc: V failed");
					return NULL;
				}
				return NULL;
			}
			if ( sem_V(queue_mutex) < 0 )
			{
				log(LOG_ERR, "get_doc: V failed");
				return NULL;
			}
			return (struct document *)-1;
		}
		else //struct present, block the process if document is not stream
		{
			if ( url->stream ) //is a stream document
			{
				if ( sem_V(queue_mutex) < 0 )
				{
					log(LOG_ERR, "get_doc: V failed");
					return NULL;
				}
				return (struct document *)-1;
			}

			REAL_ADDR(tmp)->cont++;
			if ( sem_V(queue_mutex) < 0 )
			{
				log(LOG_ERR, "get_doc: V failed");
				return NULL;
			}
			if ( sem_P(REAL_ADDR(tmp)->sem) < 0 )
			{
				log(LOG_ERR, "get_doc: P failed");
				return NULL;
			}
			if ( sem_P(queue_mutex) < 0 )
			{
				log(LOG_ERR, "get_doc: P failed");
				return NULL;
			}
			REAL_ADDR(tmp)->cont--;
			if ( sem_V(queue_mutex) < 0 )
			{
				log(LOG_ERR, "get_doc: V failed");
				return NULL;
			}
			if ( sem_P(hash_mutex) < 0 )
			{
				log(LOG_ERR, "get_doc: P failed");
				return NULL;
			}
			if ( (doc = search_doc(url->resolved_host, url->path, url->port)) == NULL ) //doc not in cache
			{
				if ( sem_V(hash_mutex) < 0 )
				{
					log(LOG_ERR, "get_doc: V failed");
					return NULL;
				}
				return (struct document *)-1;
			}
			else //doc in cache
			{
				if ( (ret = load_doc(doc)) == NULL )
					return NULL;
				if ( sem_V(hash_mutex) < 0 )
				{
					log(LOG_ERR, "get_doc: V failed");
					return NULL;
				}
				if ( sem_P(queue_mutex) < 0 )
				{
					log(LOG_ERR, "get_doc: P failed");
					return NULL;
				}
				if ( REAL_ADDR(tmp)->cont )
				{
					if ( sem_V(REAL_ADDR(tmp)->sem) < 0 )
					{
						log(LOG_ERR, "get_doc: V failed");
						return NULL;
					}
				}
				else
				{
					delete_search_dl(tmp);
				}
				if ( sem_V(queue_mutex) < 0 )
				{
					log(LOG_ERR, "get_doc: V failed");
					return NULL;
				}
				return ret;
			}
		}
	}
	else //doc in cache
	{
		ret = load_doc(doc);
		if ( sem_V(hash_mutex) < 0 )
		{
			log(LOG_ERR, "get_doc: V failed");
			return NULL;
		}
		return ret;
	}
}

//Remove host from cache.
//Return 1 on success, 0 on failure.
int remove_host(char *host)
{
	int pos;
	struct hash_host *tmp, *tmp_old;
	pos = hash_func(host);
	tmp = REAL_ADDR(hash)[pos];
	tmp_old = tmp;
	while ( (tmp != NULL) && (strcmp(REAL_ADDR(REAL_ADDR(tmp)->host_name), REAL_ADDR(host))) ) //search host
	{
		tmp_old = tmp;
		tmp = REAL_ADDR(tmp)->next;
	}
	if ( tmp == NULL ) //host not found
	{
		log(LOG_ERR, "remove_host: Host not found");
		return 0;
	}
	else if ( !strcmp(REAL_ADDR(REAL_ADDR(tmp)->host_name),
	                  REAL_ADDR(REAL_ADDR(tmp_old)->host_name)) )
		REAL_ADDR(hash)[pos] = REAL_ADDR(tmp)->next;
	else
		REAL_ADDR(tmp_old)->next = REAL_ADDR(tmp)->next;
	//free memory
	sh_free(REAL_ADDR(tmp)->host_name);
	for ( pos = 0; pos < MAXDOC; pos++ )
		if ( REAL_ADDR(tmp)->list[pos] != NULL )
		{
			sh_free(REAL_ADDR(REAL_ADDR(tmp)->list[pos])->header);
			sh_free(REAL_ADDR(REAL_ADDR(tmp)->list[pos])->path);
			sh_free(REAL_ADDR(tmp)->list[pos]);
		}
	sh_free(tmp);
	return 1;
}

//Remove the file element from cache.
//If there aren't file on the filesystem for the host we delete it.
//Return 1 on success, 0 on failure.
int remove_file(char *host, char *doc)
{
	struct hash_host *tmp;
	int i = 0, b = 1;
	tmp = find_host(host);
	if ( tmp == NULL ) //host not in cache
	{
		log(LOG_ERR, "remove_file: Host not in cache");
		return 0;
	}
	while ( (i < MAXDOC) && b ) //search doc
	{
		if ( (REAL_ADDR(tmp)->list[i] != NULL) &&
		     (REAL_ADDR(REAL_ADDR(tmp)->list[i])->path != NULL) )
			if ( !strcmp(REAL_ADDR(REAL_ADDR(REAL_ADDR(tmp)->list[i])->path),
			             REAL_ADDR(doc)) )
				b = 0;
		i++;
	}
	if ( !b )
		i--;
	if ( i == MAXDOC )
	{
		log(LOG_ERR, "remove_file: Doc not found");
		return 0;
	}
	if ( unlink(REAL_ADDR(REAL_ADDR(REAL_ADDR(tmp)->list[i])->path)) < 0 ) //remove from filesystem
	{
		log(LOG_ERR, "remove_file: Cannot delete file from fs");
		return 0;
	}
	//free memory
	sh_free(REAL_ADDR(REAL_ADDR(tmp)->list[i])->header);
	sh_free(REAL_ADDR(REAL_ADDR(tmp)->list[i])->path);
	sh_free(REAL_ADDR(tmp)->list[i]);
	REAL_ADDR(tmp)->list[i] = NULL;
	i = 0;
	while ( (i < MAXDOC) && (REAL_ADDR(tmp)->list[i] == NULL) )
		i++;
	if ( i == MAXDOC ) //there aren't file on the filesystem for the host
		if ( !remove_host(host) )
			return 0;
	return 1;
}

//Replace the document on memory and filesystem.
//Return 1 on success, 0 on failure.
int replace_doc(struct document *doc, struct urlinfo_t *url)
{
	struct hash_doc *pos;
	char *hdr;
	if ( (pos = search_doc(url->resolved_host, url->path, url->port)) == NULL ) //doc not in cache
	{
		return 0;
	}
	if ( (hdr = allocate(strlen(doc->hdr)+1, 0)) == NULL )
	{
		log(LOG_ERR, "replace_doc: Cannot allocate memory");
		return 0;
	}
	strcpy(REAL_ADDR(hdr), doc->hdr);
	sh_free(REAL_ADDR(pos)->header);

	REAL_ADDR(pos)->header = hdr; //set new header

	REAL_ADDR(pos)->size = url->size; //set new size
	if ( unlink(REAL_ADDR(REAL_ADDR(pos)->path)) < 0 ) //remove old doc from fs
	{
		log(LOG_ERR, "replace_doc: Cannot remove file from fs");
		return 0;
	}
	if ( save_doc(REAL_ADDR(REAL_ADDR(pos)->path), doc->doc, doc->doclen) == 0 ) //save doc on fs
	{
		return 0;
	}
	return 1;
}

//Clear one process blocked on a semaphore associate to url (if exist).
//Return 1 on success, 0 on failure.
int clear_one(struct urlinfo_t *url)
{
	struct search_dl *sdl;
	if ( sem_P(queue_mutex) < 0 )
	{
		log(LOG_ERR, "clear_one: P failed");
		return 0;
	}
	if ( (sdl = find_search_dl(url->resolved_host, url->path, url->port)) == NULL ) //there isn't the struct
	{
		if ( sem_V(queue_mutex) < 0 )
		{
			log(LOG_ERR, "clear_one: V failed");
			return 0;
		}
		return 1;
	}
	else //struct found
	{
		if ( REAL_ADDR(sdl)->cont ) //there are process blocked
		{
			if ( sem_V(REAL_ADDR(sdl)->sem) < 0 )
			{
				log(LOG_ERR, "clear_one: V failed");
				if ( sem_V(queue_mutex) < 0 )
				{
					log(LOG_ERR, "clear_one: V failed");
					return 0;
				}
				return 0;
			}
		}
		else //no process blocked
			delete_search_dl(sdl);
		if ( sem_V(queue_mutex) < 0 )
		{
			log(LOG_ERR, "clear_one: V failed");
			return 0;
		}
	}
	return 1;
}

//Delete from search_dl struct.
void delete_search_dl(struct search_dl *tmp)
{
	if ( tmp == NULL )
		return;
	if ( *REAL_ADDR(queue) == tmp ) //tmp is the first element
    {
		*REAL_ADDR(queue) = REAL_ADDR(tmp)->next;
		if (*REAL_ADDR(queue) != NULL)
			REAL_ADDR(*REAL_ADDR(queue))->prec = NULL;
    }
    else //tmp isn't the first element
    {
		if (REAL_ADDR(tmp)->prec != NULL)
			REAL_ADDR(REAL_ADDR(tmp)->prec)->next = REAL_ADDR(tmp)->next;
		if (REAL_ADDR(tmp)->next != NULL)
			REAL_ADDR(REAL_ADDR(tmp)->next)->prec = REAL_ADDR(tmp)->prec;
    }
	//free memory
	sh_free(REAL_ADDR(tmp)->doc);
	sh_free(REAL_ADDR(tmp)->host);
	if ( remove_sem(REAL_ADDR(tmp)->sem) < 0 ) //delete serach_dl semaphore
	{
		log(LOG_ERR, "delete_search_dl: Cannot remove search_dl semaphore");
	}
	sh_free(tmp);
}

//Search on the struct search_dl an elemento with host and doc parameter.
//Return the element if exist, NULL otherwise.
struct search_dl* find_search_dl(char *host, char *doc, int port)
{
	struct search_dl *tmp;
	int i = 1;
	tmp = *REAL_ADDR(queue);
	while ( (tmp != NULL) && i )
	{
		if ( (strcmp(REAL_ADDR(REAL_ADDR(tmp)->host), REAL_ADDR(host)) == 0)
		     && (strcmp(REAL_ADDR(REAL_ADDR(tmp)->doc), REAL_ADDR(doc)) == 0)
			 && (REAL_ADDR(tmp)->port == port) )
			i = 0; //found, exit from the loop
		else
			tmp = REAL_ADDR(tmp)->next;
	}
	return tmp;
}

//Insert (to head) an element on the search_dl struct.
//Return 1 on success, 0 on failure.
int insert_search_dl(struct urlinfo_t *url)
{
	struct search_dl *ins;
	if ( (ins = allocate(sizeof(struct search_dl),1)) == NULL ) //allocate the memory
	{
		log(LOG_ERR, "insert_search_dl: Cannot allocate memory");
		if ( sem_V(queue_mutex) < 0 )
		{
			log(LOG_ERR, "insert_search_dl: V failed");
			return 0;
		}
		return 0;
	}
	if ( (REAL_ADDR(ins)->doc = allocate(strlen(url->path)+1,1)) == NULL )
	{
		log(LOG_ERR, "insert_search_dl: Cannot allocate memory");
		sh_free(ins);
		if ( sem_V(queue_mutex) < 0 )
		{
			log(LOG_ERR, "insert_search_dl: V failed");
			return 0;
		}
		return 0;
	}
	strcpy(REAL_ADDR(REAL_ADDR(ins)->doc), url->path); //set path
	REAL_ADDR(ins)->port = url->port; //set port
	if ( (REAL_ADDR(ins)->host = allocate(strlen(url->resolved_host)+1,1)) == NULL )
	{
		log(LOG_ERR, "insert_search_dl: Cannot allocate memory");
		sh_free(REAL_ADDR(ins)->doc);
		sh_free(ins);
		if ( sem_V(queue_mutex) < 0 )
		{
			log(LOG_ERR, "insert_search_dl: V failed");
			return 0;
		}
		return 0;
	}
	strcpy(REAL_ADDR(REAL_ADDR(ins)->host), url->resolved_host); //set host
	REAL_ADDR(ins)->cont = 0;
	if ( (REAL_ADDR(ins)->sem = create_sem(0)) < 0 )
	{
		log(LOG_ERR, "insert_search_dl: Cannot create semaphore");
		sh_free(REAL_ADDR(ins)->host);
		sh_free(REAL_ADDR(ins)->doc);
		sh_free(ins);
		if ( sem_V(queue_mutex) < 0 )
		{
			log(LOG_ERR, "insert_search_dl: V failed");
			return 0;
		}
		return 0;
	}
	REAL_ADDR(ins)->next = *REAL_ADDR(queue);
	REAL_ADDR(ins)->prec = NULL;
	if (*REAL_ADDR(queue) != NULL)
		REAL_ADDR(*REAL_ADDR(queue))->prec = ins;
	*REAL_ADDR(queue) = ins; //insert in the queue
	return 1;
}

//Free the semaphores, if concurrency make it possible.
//Also removes the directory ./cache
//Return 1 on success, 0 on failure.
int finish_hash(void)
{
	struct search_dl *sdl;
	sdl = *REAL_ADDR(queue);
	while ( (sdl != NULL) && (sdl > (struct search_dl *)0) &&
	        (sdl < (struct search_dl *)(shmem_size+2)) )
	{
		if ( remove_sem(REAL_ADDR(sdl)->sem) < 0 ) //delete serach_dl semaphore
		{
			log(LOG_ERR, "finish_hash: Cannot remove search_dl semaphore");
		}
		sdl = REAL_ADDR(sdl)->next;
	}

	if ( !fork() ) //child
	{
		if ( execlp("/bin/rm", "rm", "-fR", CACHEDIR, (char *)0) < 0 ) //free filesystem
			log(LOG_ERR, "finish_hash: Cannot delete cache directory");
		exit(0);
	}
	wait(NULL);
	if ( remove_sem(hash_mutex) < 0 ) //delete hash semaphore
	{
		log(LOG_ERR, "finish_hash: Cannot remove mutex semaphore");
		return 0;
	}
	if ( remove_sem(queue_mutex) < 0 ) //delete queue semaphore
	{
		log(LOG_ERR, "finish_hash: Cannot remove mutex semaphore");
		return 0;
	}
	return 1;
}
