# include <stdio.h>
# include <stdlib.h>

# include <tree.h>
# include <list.h>
# include <logging.h>

struct node *root = NULL;
unsigned int maxdepth;

void init (struct urlinfo_t value)
{
	if ( (root = malloc (sizeof(struct node))) == NULL )
			return;
	memcpy (&(root->url),&value,sizeof(struct urlinfo_t));
	root->url.prefetch = -1;
	root->lev = 0;
	root->url.lev = 0;
	root->bros = root->son = root->father = root-> prev = NULL;
}

struct node *insert (struct urlinfo_t value,struct node *father)
{
	struct node *newnode;
	struct node *tmp;

	// Don't re-insert an existing node
	if ( (tmp=searchtree(&value,root)) != NULL )
		return tmp;

	if ( (newnode = malloc (sizeof(struct node))) == NULL)
			return (NULL);
	if ( father != NULL )
	{
		memcpy (&newnode->url,&value,sizeof (struct urlinfo_t));
		newnode->url.prefetch = 1;
		newnode->lev = (father->lev) + 1;
		newnode->url.lev = newnode->lev;
		if (newnode->lev > maxdepth)
			maxdepth = newnode->lev;
		newnode->father = father;
		newnode->bros = newnode->son = NULL;
		newnode->downloaded = 0;
		if ((tmp = father->son))
		{
			while (tmp->bros)
					tmp = tmp->bros;
			newnode->prev = tmp;
			tmp->bros = newnode;
		}
		else
		{
			newnode->prev = NULL;
			father->son = newnode;
		}
		return (newnode);
	}
	else
		return (NULL);
}

struct node *searchtree (struct urlinfo_t *value,struct node *elem)
{
	struct node *ret=NULL;
	char found = 0;

	if (elem != NULL)
	{
		if ( (strcmp(elem->url.path,value->path) == 0) &&
		     ( (strcmp(elem->url.host,value->host) == 0) ||
		       (strcmp(elem->url.resolved_host,value->resolved_host)) ) &&
			 (elem->url.port == value->port) &&
		     (strcmp(elem->url.page,value->page) == 0) )
		{
			found = 1;
			ret = elem;
		}

		if (!found)
		{
			ret = searchtree (value,elem->bros);
			if (ret == NULL)
				ret = searchtree (value,elem->son);
		}
	}
	return ret;
}

int levsearch (struct node *search,int lev)
{
	if (lev == 0)
	{
		return (0);
	}
	else if (search->lev > lev )
		return (1);
	else if ( (search->son != NULL) && (search->lev < lev) )
	{
		levsearch (search->son,lev);
	}
	else if ( (search->lev == lev) && (search->downloaded == 0) )
	{
		search->downloaded = 1;
		if ( enqueue (search->url,lev) == 0)
			log (LOG_WARNING,"levsearch: can't prefetch ");
	}
	if ( (search->bros != NULL) )
		levsearch (search->bros,lev);
	return (0);
}

void deletetree (struct node *start)
{
	if ( start != NULL)
	{
		deletetree (start->bros);
		deletetree (start->son);
		free (start);
	}
}

void updatelev (struct node *start,int newlev,int currlev)
{
	if (start != NULL)
	{
		start->lev = (start->lev - currlev) + newlev;
		start->url.lev = start->lev;
		updatelev (start->son,newlev,currlev);
		updatelev (start->bros,newlev,currlev);
	}
}

int createnewroot (struct urlinfo_t newroot)
{
	struct node *tmpnode = searchtree (&newroot,root) ;

	if (tmpnode) // node is in the tree, delete only older things
	{
		tmpnode->url.prefetch = -1;
		tmpnode->prev->bros = NULL;
		tmpnode->prev = NULL;
		deletetree (root);
		root = tmpnode->bros;
		tmpnode->bros->prev = NULL;
		tmpnode->bros = NULL;
		deletetree (root);
		root = tmpnode;
		updatelev (root,0,root->lev);
		return (1);
	}
	else // create a new network tree
	{
		newroot.prefetch = -1;
		deletetree (root);
		init (newroot);
		return (1);
	}
	return (-1); // never reached
}

struct node *delgtree (struct urlinfo_t *value,struct node *elem)
{
	/* when delting a node, delete the subtree too */
	struct node *tmp;

	tmp = searchtree (value,elem);
	if (tmp->bros != NULL)
	{
		if (tmp->prev != NULL)
		{
			tmp->prev->bros = tmp->bros;
			tmp->bros->prev = tmp->prev;
		}
		else
		{
			tmp->father->son = tmp->bros;
		}
	}
	deletetree (tmp);
	return tmp;
}

void visita(struct node *start)
{
	if (start != NULL)
	{
		visita (start->bros);
		visita (start->son);
	}
}

