#include <stdlib.h>
#include <string.h>
#include <conv_date.h>
#include <logging.h>

static char *month_ptr, *year_ptr, *time_ptr;
static int format;

// get the day of the week
char* get_dow(struct tm* data) {
	char* dow=NULL;

	switch(data->tm_wday) {
		case 0: dow="Sun, "; break;
		case 1: dow="Mon, "; break;
		case 2: dow="Tue, "; break;
		case 3: dow="Wed, "; break;
		case 4: dow="Thu, "; break;
		case 5: dow="Fri, "; break;
		case 6: dow="Sat, "; break;
	}
	return dow;
}

// convert to_conv into a string of the specified length (n_char)
char* int_to_str(int to_conv, int n_char) {
	char *stringa;
	int cifra, cont=0;

	if ((stringa=malloc(n_char+1))==NULL) {
		log(LOG_ERR,"int_to_str: Not enough memory");
		return NULL;
	}
	stringa[n_char]='\0';
	while (n_char>0) {
		cifra=to_conv%10;
		to_conv-=cifra;
		to_conv/=10;
		cifra+=48;
		stringa[n_char-1]=(char)cifra;
		cont++;
		n_char--;
	}
	return stringa;
}

char* get_month(struct tm* data) {
	char* month=NULL;

	switch(data->tm_mon) {
		case  0: month=" Jan "; break;
		case  1: month=" Feb "; break;
		case  2: month=" Mar "; break;
		case  3: month=" Apr "; break;
		case  4: month=" May "; break;
		case  5: month=" Jun "; break;
		case  6: month=" Jul "; break;
		case  7: month=" Aug "; break;
		case  8: month=" Sep "; break;
		case  9: month=" Oct "; break;
		case 10: month=" Nov "; break;
		case 11: month=" Dec "; break;
	}
	return month;
}

// it returns the date in the format specified by RFC 1123, NULL on error
char* conv_date(struct tm* data) {

	char* string_date;
	char* tmp;

	if (data==NULL) {
		return NULL;
	}

	if ((string_date=malloc(STRING_LEN))==NULL) {
		log(LOG_ERR,"conv_data: Not enough memory");
		return NULL;
	}
	string_date[0]='\0';

	strncat(string_date, get_dow(data), 5);

	if ((tmp=int_to_str(data->tm_mday, 2))==NULL) {
		log(LOG_ERR,"conv_data: Not enough memory");
		free(string_date);
		return NULL;
	}
	strncat(string_date, tmp, 2);
	free(tmp);

	strncat(string_date, get_month(data), 5);

	if ((tmp=int_to_str((data->tm_year)+1900, 4))==NULL) {
		log(LOG_ERR,"conv_data: Not enough memory");
		free(string_date);
		return NULL;
	}

	strncat(string_date, tmp, 4);
	strncat(string_date, " ", 1);

	if ((tmp=int_to_str((data->tm_hour), 2))==NULL) {
		log(LOG_ERR,"conv_data: Not enough memory");
		free(string_date);
		return NULL;
	}
	strncat(string_date, tmp, 2);
	strncat(string_date, ":", 1);

	if ((tmp=int_to_str((data->tm_min), 2))==NULL) {
		log(LOG_ERR,"conv_data: Not enough memory");
		free(string_date);
		return NULL;
	}
	strncat(string_date, tmp, 2);
	strncat(string_date, ":", 1);

	if ((tmp=int_to_str((data->tm_sec), 2))==NULL) {
		log(LOG_ERR,"conv_data: Not enough memory");
		free(string_date);
		return NULL;
	}
	strncat(string_date, tmp, 2);
	strncat(string_date, " GMT", 4);
	
	return string_date;
}

// insert into the tm structure the right day of the week
void parse_dow(char* dow_ptr, struct tm* data) {
	char tmp_ptr[4];

	strncpy(tmp_ptr, dow_ptr, 3);
	tmp_ptr[3]='\0';
	
	if (strcmp(tmp_ptr, "Mon")==0) data->tm_wday=1;
	else if (strcmp(tmp_ptr, "Tue")==0) data->tm_wday=2;
	else if (strcmp(tmp_ptr, "Wed")==0) data->tm_wday=3;
	else if (strcmp(tmp_ptr, "Thu")==0) data->tm_wday=4;
	else if (strcmp(tmp_ptr, "Fri")==0) data->tm_wday=5;
	else if (strcmp(tmp_ptr, "Sat")==0) data->tm_wday=6;
	else if (strcmp(tmp_ptr, "Sun")==0) data->tm_wday=0;
	else data->tm_wday=7;

	switch (dow_ptr[3]) {
		case ',': format=1; break;
		case ' ': format=3; break;
		default : format=2; break;
	}
}

// insert into the tm structure the right day of the month
void parse_dom(char* data_ptr, struct tm* data) {
	char dom_ptr[3], *tmp_data;

	tmp_data=data_ptr;
	while ((*tmp_data)<'0'||(*tmp_data)>'9') 
		tmp_data++;
	strncpy(dom_ptr, tmp_data, 2);

	if (dom_ptr[1]==' ')
		dom_ptr[1]='\0';
	else dom_ptr[2]='\0';

	data->tm_mday=atoi(dom_ptr);

	if (data->tm_mday>31)
		data->tm_mday=0;

	if (format==3)
		month_ptr=&(data_ptr[4]);
	else month_ptr=(tmp_data+3);
}

// insert into the tm structure the right month
void parse_month(char* data_ptr, struct tm* data) {
	char tmp_ptr[4];

	strncpy(tmp_ptr, month_ptr, 3);
	tmp_ptr[3]='\0';

	if (strcmp(tmp_ptr, "Jan")==0) data->tm_mon= 0;
	else if (strcmp(tmp_ptr, "Feb")==0) data->tm_mon= 1;
	else if (strcmp(tmp_ptr, "Mar")==0) data->tm_mon= 2;
	else if (strcmp(tmp_ptr, "Apr")==0) data->tm_mon= 3;
	else if (strcmp(tmp_ptr, "May")==0) data->tm_mon= 4;
	else if (strcmp(tmp_ptr, "Jun")==0) data->tm_mon= 5;
	else if (strcmp(tmp_ptr, "Jul")==0) data->tm_mon= 6;
	else if (strcmp(tmp_ptr, "Aug")==0) data->tm_mon= 7;
	else if (strcmp(tmp_ptr, "Sep")==0) data->tm_mon= 8;
	else if (strcmp(tmp_ptr, "Oct")==0) data->tm_mon= 9;
	else if (strcmp(tmp_ptr, "Nov")==0) data->tm_mon=10;
	else if (strcmp(tmp_ptr, "Dec")==0) data->tm_mon=11;
	else data->tm_wday=7;

	if (format==3)
		year_ptr=&(data_ptr[20]);
	else year_ptr=&(month_ptr[4]);
}

// insert into the tm structure the right year 
void parse_year(char* data_ptr, struct tm* data) {
	char tmp_year[5];

	strncpy(tmp_year, year_ptr, 4);
	if (format==2) 
		tmp_year[2]='\0';
	else tmp_year[4]='\0';

	data->tm_year=atoi(tmp_year);
	if (format==2) {
		time_t* tempo;
		char *serv_data;
		int my_year;

		tempo=malloc(sizeof(time_t));
		if (tempo!=NULL) {
			time(tempo); /* the time since the Epoch (00:00:00 UTC, January 1, 
						    1970), measured in seconds, is stored in "tempo" */
			serv_data=ctime(tempo); /* the ctime() function converts "tempo" 
									   into a string of the form: 
									   "Wed Jun 30 21:49:08 1993" */
			serv_data+=YEAR_OFFSET;
			my_year=atoi(serv_data);

			time_ptr=year_ptr+3;
			data->tm_year+=2000;
			if ((data->tm_year-my_year)>50)
				data->tm_year-=100;
			free(tempo);
		}
		else {
			data->tm_wday=7;
			log(LOG_ERR,"parse_year: Not enough memory");
		}
	}
	else if (format==1) time_ptr=year_ptr+5;
	else time_ptr=&(data_ptr[11]);

	// Year must be stored as number of years from 1900
	data->tm_year -= 1900;
}

// parses time_ptr, inserting insert into the tm structure
// the right hours, minutes and seconds 
void parse_time(struct tm* data) {
	int cont;
	char tmp_time[3];

	tmp_time[2]='\0';
	for (cont=1; cont<4; cont++) {
		strncpy(tmp_time, time_ptr, 2);
		time_ptr+=3;
		if (cont==1) data->tm_hour=atoi(tmp_time);
		else if (cont==2) data->tm_min=atoi(tmp_time);	
		else if (cont==3) data->tm_sec=atoi(tmp_time);	
	}
	if (data->tm_hour< 0||data->tm_min< 0||data->tm_sec< 0||  \
	    data->tm_hour>24||data->tm_min>59||data->tm_sec>59)
		data->tm_wday=7;
}

/* passing to this function a date (in 1 of the 3 formats admitted), it returns
   a tm structure; on error the tm_wday field is set to the invalid value 7 */
struct tm parse_date(char* data_ptr) {
	struct tm data;
	
	parse_dow(data_ptr, &data);
	if (VALID(data)) {
		parse_dom(data_ptr, &data);
		if (VALID(data)) {
			parse_month(data_ptr, &data);
			if (VALID(data)) {
				parse_year(data_ptr, &data);
				if (VALID(data)) 
					parse_time(&data);
			}
		}
	}
	return data;
}
