#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include <hdr_parser.h>
#include <buffers.h>
#include <logging.h>

int add_header(char **hdr, char *ins, int check_data)
{
	char *datalen_c;
	int datalen;

	int start_data;
	char backup;
	char *end_hdr;
	char *new_hdr;

	if ( (end_hdr=get_http_doc_start(*hdr)) == NULL )
	{
		log(LOG_ERR,"add_header: Not a valid header");
		return 0;
	}

	if (check_data)
	{
		// Get the length of appended data, if present
		// Here we don't support "Transfer-encoding: chunked", but there's
		// no need of doing it, because we will handle here only headers
		// with Content-Length or nothing.
		if ( (datalen_c=parse_http_hdr("Content-Length",*hdr)) == NULL )
		{
			datalen = 0;
		}
		else
		{
			datalen = atoi(datalen_c);
			free(datalen_c);
		}
	}
	else
	{
		datalen = 0;
	}

	backup=*(end_hdr-2);
	*(end_hdr-2)='\0';

	if ( (new_hdr=malloc(strlen(*hdr)+strlen(ins)+datalen+5)) == NULL )
	{
		*(end_hdr-2)=backup;
		log(LOG_ERR,"add_header: Not enough memory");
		return 0;
	}

	strcpy(new_hdr,*hdr);
	strcat(new_hdr,ins);
	strcat(new_hdr,"\r\n\r\n");

	start_data=strlen(new_hdr);
	memcpy(&new_hdr[start_data],end_hdr,datalen);
	new_hdr[start_data+datalen]='\0';

	*(end_hdr-2)=backup;
	free(*hdr);
	*hdr=new_hdr;

	return 1;
}

int delete_header(char **hdr, char *del)
{
	char *tmp_hdr;
	char *tmp;
	char *del_safe;

	if ( (del_safe=malloc(strlen(del)+2)) == NULL )
	{
		log(LOG_ERR,"delete_header: Not enough memory");
		return 0;
	}
	strcpy(del_safe,"\n");
	strcat(del_safe,del);

	if ( (tmp=strstrcase(*hdr,del_safe)) == NULL )
	{
		//log(LOG_WARNING,"delete_header: Cannot find header");
		// Even if we get here it's fine, so we return 1.
		free(del_safe);
		return 1;
	}
	free(del_safe);
	tmp++;

	if ( (tmp_hdr=malloc(strlen(*hdr)+1)) == NULL )
	{
		log(LOG_ERR,"delete_header: Not enough memory");
		return 0;
	}
	memset(tmp_hdr,0,strlen(*hdr)+1);

	strncpy(tmp_hdr,*hdr,tmp-(*hdr));

	do
	{
		tmp++;
	}
	while ( (*tmp != '\0') && ( (*(tmp-1) != '\n') || (is_sptab(*tmp)) ) );

	strcat(tmp_hdr,tmp);

	free(*hdr);

	if ( (*hdr=malloc(strlen(tmp_hdr)+1)) == NULL )
	{
		/* Cannot save memory space... just return tmp_hdr, which is
		 * shorter than occupied space, but can be good
		 */

		*hdr=tmp_hdr;
		return 1;
	}

	strcpy(*hdr,tmp_hdr);
	free(tmp_hdr);

	return 1;
}

int modify_header_except_TE(char **header)
{
	char *ptr;

	if (!delete_header(header,"Proxy-Authenticate"))
	{
		return 0;
	}
	if (!delete_header(header,"Proxy-Authorization"))
	{
		return 0;
	}
	if (!delete_header(header,"TE"))
	{
		return 0;
	}
	if (!delete_header(header,"Trailers"))
	{
		return 0;
	}
	if (!delete_header(header,"Upgrade"))
	{
		return 0;
	}

	if ( (ptr=parse_http_hdr("Connection",*header)) != NULL )
	{
		char *back;
		char *forward;

		back=forward=ptr;

		while (*forward != '\0')
		{
			while ( (*forward != '\0') && (!is_sptabcomma(*forward)) )
				forward++;

			if (*forward != '\0')
			{
				*forward='\0';
				if (!delete_header(header,back))
				{
					free(ptr);
					return 0;
				}
				forward++;
				while ( (*forward != '\0') && (is_sptab(*forward)) )
					forward++;
				back=forward;
			}
			else
			{
				if (!delete_header(header,back))
				{
					free(ptr);
					return 0;
				}
			}
		}
		free(ptr);

		if (!delete_header(header,"Connection"))
		{
			return 0;
		}
	}

	return 1;
}

int modify_header(char **header)
{
	if (!modify_header_except_TE(header))
	{
		return 0;
	}
	if (!delete_header(header,"Transfer-Encoding"))
	{
		return 0;
	}

	return 1;
}

int add_via_header(char **request)
{
	char hname[256];
	char *old_via;
	char *field;

	if ( gethostname (hname,sizeof(hname)) == -1 )
		strcpy (hname,"unknown-" NAME);
	hname[sizeof(hname)-1]='\0';

	old_via=parse_http_hdr("Via", *request);

	if (old_via == NULL)
	{
		if ( (field=malloc(strlen(NAME)+strlen(VERSION)+
		                   strlen(hname)+14)) == NULL )
		{
			log(LOG_ERR,"add_via_header: Not enough memory");
			return 0;
		}
		sprintf(field,"Via: 1.1 %s (" NAME "/" VERSION ")",hname);
	}
	else
	{
		if (!delete_header(request,"Via"))
		{
			log(LOG_ERR,"add_via_header: Cannot modify header");
			free(old_via);
			return 0;
		}

		if ( (field=malloc(strlen(NAME)+strlen(VERSION)+
		                   strlen(old_via)+strlen(hname)+16)) == NULL )
		{
			log(LOG_ERR,"add_via_header: Not enough memory");
			free(old_via);
			return 0;
		}
		sprintf(field,"Via: %s, 1.1 %s (" NAME "/" VERSION ")",old_via,hname);
		free(old_via);
	}

	if (!add_header(request,field,1))
	{
		log(LOG_ERR,"add_via_header: Cannot modify header");
		free(field);
		return 0;
	}

	free(field);
	return 1;
}


