# include <stdio.h>
# include <string.h>
# include <stdlib.h>
# include <unistd.h>
# include <ctype.h>

# include <hrefparse.h>
# include <urlparse.h>
# include <tree.h>
# include <logging.h>
# include <struct.h>
# include <const.h>

struct urlinfo_t baseurl;
char basehost [MAX_URL];

int definebasehost (const char *host)
{
	strcpy (basehost,host);
	return (1);
}

void shift (char *path)
{
	int fst = 0;
	int snd = 1;
	
	while ( path [snd] == '\0')
	{
		path [fst] = path [snd];
		fst++;
		snd++;
	}
	path [fst] = '\0';
}

/* fill in the struct which keep information about page absolute path */
int fillinbase (struct urlinfo_t url)
{
	char *del_pag_name;
	int pathlen = 0;

	bzero (&baseurl,sizeof(struct urlinfo_t));
	strcpy(baseurl.proto,url.proto);
	strcpy(baseurl.host,url.host);
	definebasehost (url.host);
	strcpy(baseurl.resolved_host,url.resolved_host);
	strcpy(baseurl.path,url.path);
	pathlen = strlen (baseurl.path);
	if (pathlen < (MAX_PATH_LEN-2))
	{
		if ( baseurl.path[pathlen-1] != '/' )
		{
			baseurl.path [pathlen] = '/';
			baseurl.path [pathlen+1] = '\0';
		}
	}
	if ( (baseurl.path [0] == '/') && (baseurl.path [1] == '/') )
	{
		shift (baseurl.path);
	}

	if ( (del_pag_name=strstr(baseurl.path,"htm")) != NULL )
	{
		while ( (*del_pag_name != '/') && (del_pag_name != baseurl.path) )
			del_pag_name --;
		del_pag_name++;
		strcat (baseurl.page,del_pag_name);
		*del_pag_name = '\0';
		if (del_pag_name == (baseurl.path+1))
		{
			baseurl.path[0] = '/';
			baseurl.path[1] = '\0';
			baseurl.page[0] = '\0';
		}
	}
	strcpy(baseurl.query,url.query);
	baseurl.port = (url.port != 0) ? url.port : 80;
	baseurl.stream = 0;
	baseurl.fragment = 0;
	baseurl.prefetch = -2;
	baseurl.size = url.size;
	return (1);
}

int ttoi (const char *type)
{
	if (!strcmp(type,BASE))
		return (0);
	if (!strcmp(type,A))
		return (1);
	return (-1);
}

/*
 * search for all hyperlinks in a page. If found insert them in a general tree and create 
 * a prediction network
 */

int tagparser (const char *page,struct node *father,const char *type,
               const char *tag)
{
	char *search;
	char *newpage;
	char tmpurl[MAX_URL];
	struct urlinfo_t urlinfo;
	int count;
	int ntype=-1;

	newpage = search = (char *)malloc(strlen(page)+1);	
	strcpy (search,page); ntype = ttoi (type);
	while (*search != '\0')
	{
		int found=0;

		while ( (strncasecmp (search,type,strlen(type)) != 0) && \
			    (*search != '\0') )
		{
			search ++;
		}
		if (strncasecmp(search,type,strlen(type)) == 0)
			found = 1;

		if (found)
		{
			// anchor or base tag founded
			search += strlen (type);
			// now search for an href tag which specify an address
			while ( (strncasecmp (search,tag,strlen(tag)) != 0) && \
					(*search != '\0') )
			{
				search++;
			}
		}

		if ( (found) && (strncasecmp (search,tag,strlen(tag)) == 0) ) // href found
		{
			search += strlen(tag);
			/* goes to start of address */
			while ( (*search != '=') && (*search != '\0') )
				search++;
			while ((!isalnum(*search)) && (*search != '"') && (*search != '\0'))
				search++;
			if (*search == '"') search++;

			count = 0;
			while ( ((*search) != '"') && ((*search) != '\0') &&
			        ((*search) != ' ') && ((*search) != '\n') &&
			        ((*search) != '\r') && ((*search) != '>') &&
			        (count < MAX_URL) )
				tmpurl[count++] = (*(search++));
	    	if (count >= MAX_URL)
					tmpurl[count-1] = '\0';
			else
					tmpurl[count] = '\0';

			if (urlParse (tmpurl,&urlinfo,DEFAULT) >= 0)
			// divide and fill in struct urlinfo_t
			{
				if (strncmp(type,BASE,strlen(BASE)) == 0) // if a base tag is founded, fill in the base struct
				{
					fillinbase(urlinfo);
				}
				else // if the link is relative, we must insert host and the first part of the path
				{
					if (urlinfo.host[0] == '\0') 
					{
						if (baseurl.host[0] == '\0') 
							strcpy (urlinfo.host,basehost);
						else
						{
							strcpy (urlinfo.host,baseurl.host);	
							strcpy (urlinfo.resolved_host,baseurl.resolved_host);	
						}
					}
					isstream (&urlinfo); // check if resources pointed is stream o lead

				/* insert a new node in the prediction network tree */
					if (urlinfo.fragment == 1)
					{
						if ( !(searchtree (&urlinfo,root)) )
						{
							if ( insert (urlinfo,father) == NULL)
							{
								log (LOG_WARNING,"hrefparse: can't prefetch");
							}
						}
					}
					else
					{
						if ( insert (urlinfo,father) == NULL )
						{
							log (LOG_WARNING,"hrefparse: can't prefetch");
						}
					}
				}
			}
		}
		if ( strcasecmp (search,BASE) == 0 )
		{
			free(newpage);
			return (1);
		}
	}
	free(newpage);
	return (-1);
}

/* this is the function called by proxy for prefetching */

int searchforhref (char *page,struct urlinfo_t url)
{
	struct node *father;
	fillinbase(url);
	if (root == NULL)
		init(url);
	father = searchtree (&url,root);
	if (father == NULL)
		father = root;
	tagparser(page,father,BASE,HREF);
	tagparser(page,father,IMG,SRC);
	tagparser(page,father,A,HREF);
	return (maxdepth);
}

