#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <logging.h>
#include <struct.h>
#include <baseurl.h>
#include <const.h>

/* files not prefetched (size too big) */
const char *extension[] = {
	"avi",
	"mpg",
	"mpeg",
	"mp3",
	"gz",
	"tar",
	"rar",
	"bz2",
	"Z",
	"zip",
	"tgz",
	"exe"
};

char *extractpath (char *tmpurl,struct urlinfo_t *ret,int relative);
char *extracthost (char *tmpurl,struct urlinfo_t *ret,int mode,int relative);
int isstream (struct urlinfo_t *ret);

/*
 * compress '//' into '/'
 */

void compress_slash(char *string)
{
	int i=0;
	while (string[i] != '\0')
	{
		if ( (string[i] == '/') && (string[i+1] == '/') )
		{
			memmove(&string[i],&string[i+1],strlen(&string[i+1])+1);
		}
		else
			i++;
	}
}

/*
 * convert %HEXHEX notation in human readable format
 * if allocate == 1 new pointer to converted string (old and new) 
 */

char * url_convert_hex(char *org_url, int allocate)
{
    static char code[] = "00";
    char *url = NULL;
    char *s = NULL;
	char *t = NULL;
    url = allocate ? (char *) strdup(org_url) : org_url;
    if ((int) strlen(url) < 3 || !strchr(url, '%'))
	    return url;
    for (s = t = url; *(s + 2); s++) 
	{
	    if (*s == '%') 
		{
	        code[0] = *(++s);
	        code[1] = *(++s);
	        *t++ = (char) strtol(code, NULL, 16);
	    } 
		else 
			*t++ = *s;
    }
    do {
	    *t++ = *s;
    } while (*s++);
    return url;
}

int defaultPort (char protocol[])
{
	if (strcmp (protocol,"http") == 0)
		return (80);
	if (strcmp (protocol,"ftp") == 0)
		return (21);
	if (strcmp (protocol,"news") == 0)
		return (110);
	if (strcmp (protocol,"file") == 0)
		return (-1); 
	return (80);
}

int urlParse (char *url, struct urlinfo_t *ret,int mode)
{
	char *sharp_ptr;
	char *tmpurl=url;
	int count=0;
	int relative=0;
	int lenpath=0;
	
	bzero (ret,sizeof (struct urlinfo_t));	
	
	if (strchr(tmpurl,'@'))
		return (-1);
	if (strncasecmp(tmpurl,"ftp://",6) == 0)
		return (-1);
	if (strncasecmp(tmpurl,"news://",7) == 0)
		return (-1);
	if ( (sharp_ptr=strrchr(tmpurl,'#')) != NULL )
	{
		if (*tmpurl != '#')
		{
			tmpurl[0]='.';
			tmpurl[1]='/';
			tmpurl[2]='\0';
			ret->fragment = 1;
		}
		else
		{
			*sharp_ptr = '\0';
			ret->fragment = 2;
		}
		relative = 1;
	}
	else
		ret->fragment = 0;

	if ( strchr (url,'%') )
		tmpurl = url_convert_hex (url,1);
	/* if the char ':' is not present and the url you are examining is
	 * a href in an html page, then the path is not absolute, because it is 
	 * not in the form <protocol>:<host>.... or in the form <host>:<port>.
	 * it could be <host>/<path>, but i think it could generate misunderstending
	 * so...
	 */ 
	if ( (*tmpurl == '.') || (*tmpurl == '/') || \
		 ((strchr (url,':') == NULL)) )
		relative = 1;
	if ( (*tmpurl == '.') && (*(tmpurl+1) == '/') )
	{
		tmpurl +=2;
	}
	while ( ((*tmpurl) != ':') && ((*tmpurl) != '\0') && (relative != 1) && (count<MAX_PROTO_LEN) )
		ret->proto[count++] = *tmpurl++;
	if ( (*tmpurl == '\0') || (relative == 1) ) 
	{
		// if addr is relative, information are taken from baseurl
		strcpy(ret->proto,baseurl.proto);
		ret->proto[strlen(baseurl.proto)+1] = '\0';
		tmpurl = url;
	}
	else
	{
		ret->proto[count] = '\0';
	}

	count = 0;
	while ( (*(++tmpurl) == '/') && (count < 2) && (relative != 1) )
		count++;
	if ( relative == 1 )
	{
		count = 2;
		tmpurl -= 1;
	}
	if ( (count != 2) && (strncmp (ret->proto,"http",4) == 0) )
		return (-1); // Malformed url

	if ( (tmpurl = extracthost(tmpurl,ret,mode,relative)) == 0)
		return (-1);

	if ( (tmpurl = extractpath (tmpurl,ret,relative)) == 0) 
		return (-1);

	lenpath = strlen (ret->path);

	if ( (ret->path[lenpath-1] != '/') && (lenpath < (MAX_PATH_LEN-2)) &&
	     (strchr(ret->path,'.') == NULL) && (strchr(ret->path,'#') == NULL) )
	{
		ret->path [lenpath] = '/';
		ret->path [lenpath+1] = '\0';
	}
	compress_slash(ret->path);
	isstream (ret);

	return (1);
}

/* extract host from url */
char *extracthost (char *tmpurl,struct urlinfo_t *ret,int mode,int relative)
{
	int count = 0;
	int nport = 0;
	char *endptr[2];

	if (relative != 1)
	{
		while ( ((*tmpurl) != ':') && ((*tmpurl) != '/') && ((*tmpurl) != '\0') && 
				(count < MAX_HOST_LEN) )
			ret->host [count++] = *tmpurl++;
		
	    if ( *tmpurl == ':' ) // is specified a port different from default, take it
	    {
		    tmpurl++;
		    nport = strtol ((const char *)tmpurl,endptr,10);
		    if (*endptr[0] != '/')
			    return (NULL); //malformed url
			tmpurl = endptr[0];
		}
		
		if (*tmpurl == '/') // delete the / at the end of the hostname.
		{
			ret->host[count] = '\0';
		}
		else
		{
			return (NULL); // malformed url
		}
	
		count=0;
	}
	else
	{
        strcpy (ret->host,baseurl.host);
        strcpy (ret->resolved_host,baseurl.resolved_host);
	}
	ret->port = (nport != 0) ? nport : (defaultPort (ret->proto));

	return (tmpurl);
}

char *find_prec_dir (char *tmpurl,struct urlinfo_t *url)
{
	int prec_dir_num = 0;
	int path_len = 0;

	path_len = strlen (baseurl.path) - 2;
	while ( strncmp(tmpurl,"../",3) == 0  )
	{
		prec_dir_num ++;
		tmpurl += 3;
	}
	strcpy (url->path,baseurl.path);
	while ( (path_len > 0) && (prec_dir_num > 0) )
	{
		prec_dir_num --;
		while (url->path[path_len] != '/')
			path_len --;
		url->path[path_len+1] = '\0';
	}
	if (prec_dir_num < 0)
	{
		log (LOG_WARNING,"prec_dir: too many ../");
		*tmpurl ='/';
		return (tmpurl);
	}
	return (tmpurl);
}

/* extract path from url */
char *extractpath (char *tmpurl,struct urlinfo_t *ret,int relative)
{
	int count = 0;
	int toplevel = 0;
	int len = 0;
	
	if ( strncmp (tmpurl,"..",2) == 0 )
	{
		tmpurl = find_prec_dir (tmpurl,ret);
		relative = 0;
	}
	if ((*tmpurl) == '.')
		tmpurl++;
	else if ((*tmpurl) == '/')
		toplevel=1;

	if (relative) // relative path, abs path format is: /page_host_path/relative_path
	{
		if (!toplevel)
		{
			while ( baseurl.path[count] != '\0' )
				ret->path[count] = baseurl.path[count++];
			if ( (ret->path [count-1] == '/') && (strlen (baseurl.page) != 0) ) 
				ret->path [--count] = '\0';
		}
		if (isalnum(*tmpurl))
		{
			ret->path[count++]='/';
			ret->path[count++]='\0';
		}
		if (ret->fragment == 2)
		{
			strcat (ret->path,baseurl.page);
			count = strlen (ret->path) + 1;
			ret->fragment = 1;
		}
	}
	else // if addr doesn't start with /, we insert it (used for GET request)
	{
		if ( ((*tmpurl) != '/') && (ret->path[count-1] != '/') )
			ret->path[count++] = '/';
	}
	if ( (ret->path[0] == '/') && (ret->path[1] == '\0') && (*tmpurl == '/') )
		ret->path[0] = '\0';

	if (strchr(tmpurl,'?')) // search for query
	{
		while ( (*tmpurl) != '?' )
			ret->path[count++] = *tmpurl++;
		ret->path[count] = '\0';
		if (ret->path[0] == '\0')
		{
			ret->path[0]='/';
			ret->path[1]='\0';
		}
		tmpurl++;
		count = 0;
		while ( (*tmpurl) != '\0' )
			ret->query[count++] = *tmpurl++;
		ret->query[count]='\0';
	}
	else // no query found
	{
		int len = strlen (ret->path);

		if ( (ret->path [len-1] == '/') && (*tmpurl == '/') )
			ret->path [len-1] = '\0';
		if ( (strlen(tmpurl) > 1) || (*tmpurl != '/') )
			strcat (ret->path,tmpurl);
		bzero (ret->query,MAX_QUERY_LEN);
	}
	if ( strncmp(ret->path,"/",1) == 0)
		return (tmpurl);
	
	if ( (len=strlen (ret->path)) > 0 )
	{
		if ( (ret->path [len-1] != '/') && (len < 255) )
		{
			ret->path[len-1] = '/';
			ret->path[len] = '\0';
		}
	}
	else
	{
		ret->path [0] = '/';
		ret->path [1] = '\0';
	}
	return (tmpurl);
}

/* check if resource is a stream */
int isstream (struct urlinfo_t *ret)
{
	char *ptr;
	char ext[5];
	int i;

	if ( (ret->path[0]) != '\0' )
	{
		if ( (ptr=strrchr(ret->path,'.')) != NULL )
		{
			ptr++;
			i=0;

			while ( (i<sizeof(ext)-1) && (*(ptr+i) != '#') &&
			     (*(ptr+i) != '?') && (*(ptr+i) != '\0') )
			{
				ext[i]=*(ptr+i);
				i++;
			}
			ext[i]='\0';
		}
		else
			strcpy (ext,"htm");
		ret->stream = 0;
		for (i = 0;i < MAX_NUM_EXT;i++)
		{
			if ( strcasecmp (ext,extension[i]) == 0 )
			{
				ret->stream = 1;
				return (1);
			}
		}
	}
	return (0);
}
