#include <proxy_main.h>
#include <list.h>
#include <struct.h>
#include <const.h>
#include <tree.h>
#include <prefetch.h>
#include <logging.h>

/* this function prepare a request, formed by a certain number of streaming 
 * resources, accordingly with max bandwidth for streaming channel
 */

int lead_maxdim;
int stream_maxdim;

unsigned int mtu_return (int interface)
{
	if (interface == PPP)
		return (552);
	if (interface == ETH)
		return (1500);
	return (0);
}

int createqueue (int start,int maxlev)
{
	while ( (start <= maxlev+1) && (start <= maxdepth) )
	{
		/* create an ordered queue from which agent can have request for prefetching */
		levsearch(root,start);
		start++;
	}
	return (1);	
}

int send_req (int interface)
{
	int lev = 1;

	if (!head)
	{
		if ( !initqueue ())
		{
			log (LOG_ERR,"send_req: can't start prefetching");
			return (0);
		}
	}
	lead_maxdim = (mtu_return (interface) / 100) * MAX_LEAD_PERCENT;
	stream_maxdim = mtu_return(interface) - lead_maxdim;
	createqueue (lev,MAX_PREFETCH_LEVEL);
	return (1);
}

int obtain_stream ()
{
	int stream_dim = 0;
	struct hrefurl *tmp;

	if ((stream_head) && (stream_head->first))
	{
		while ( (stream_dim < stream_maxdim) && (stream_head->first != NULL) )
		{
			tmp = extract_stream ();
			if (tmp->url->stream == 1)
				start_prefetch (*(tmp->url));
			stream_dim += tmp->url->size;
			dequeue (tmp);
		}
		return (1);
	}
	else
	{
		log (LOG_WARNING,"obtain_lead: lead queue is empty (no links on page?)");
		return(0); // lead queue is empty, if bandwidth is available, restart prefetch
	}
	return (-1); //never reached
}

int obtain_lead ()
{
	int lead_dim = 0;
	struct hrefurl *tmp;

	if ( (lead_head) && (lead_head->first) )
	{
		// lead_dim check can be useful for bandwith management, but we
		// don't use it.
//		while ( (lead_dim < lead_maxdim) && (lead_head->first != NULL) )
		while (lead_head->first != NULL)
		{
			tmp = extract_lead ();
			if (tmp->url->stream == 0)
				start_prefetch (*(tmp->url));
			lead_dim += tmp->url->size;
			dequeue (tmp);
		}
		return (1);
	}
	else
	{
		log (LOG_WARNING,"obtain_lead: lead queue is empty (no links on page?)");
		return(0); // stream queue is empty, if bandwidth is available, restart prefetch
	}
	return (-1); //never reached
}

int increase_level (int addlev)
{
	/* if channel is free, you can decide to continue prefetch or you can
	 * start to download the
	 * stream part. You can decide to call obtain_stream from here
	 */

	int max = 0;

	if (stream_head->u.retr_level > lead_head->u.retr_level) 
		max += stream_head->u.retr_level; 
	else
		max += lead_head->u.retr_level;
	addlev += max;
	createqueue (max+1,addlev);
	obtain_lead ();
	return (1);
}
