#include <stdlib.h>

#include <init_shared.h>

#define ABS(x) ((x>=0) ? (x) : -(x))

/* Type of the header of each allocated block */
#define HTYPE int
#define HEADER_SIZE sizeof(HTYPE)

/* Value of the header of the last block of the shared memory */
#define LAST_SHMEM 0

/* Conversion of pointer to HTYPE */
#define HPTR *(HTYPE *)

/* pointer to the last byte of shared memory */
#define SHMEM_END (shmem_start+shmem_size)

#ifdef DEBUG_SHDUMP
#include <stdio.h>
void sh_dump(int size)
{
	int i;
	printf("Dumping first %d bytes of shared memory...\n",size);
	for (i=0;i<size;i++)
		printf("%.2X ",*((unsigned char*)(shmem_start+i)));
	printf("\n");
}
#endif

// this function returns a pointer to an area of available shared memory,
// if no memory is available, NULL is returned
void *sh_malloc(int memsiz) 
{
	char *ptr,*cursor;
	int totsiz;
	HTYPE tmp;
	int found, end, tmp_flag;

	/* If initial allocation was bad exits immediatly */
	if (shmem_size<=0)
	{
		return NULL;
	}

	found=end=0;
	cursor=ptr=shmem_start;

	while (  ( !end ) && ( ( SHMEM_END - cursor ) > 0 )  )
	{
		ptr = cursor;

		/* Discard all the allocated and still used blocks */
		tmp_flag=0;
		while  ( ( ( SHMEM_END - ptr ) > 0 ) && ( !tmp_flag ) )
		{
			if ( (tmp=(HPTR ptr)) > 0 )
			{
				ptr += ABS(tmp);
				ptr += HEADER_SIZE;
			}
			else
			{
				tmp_flag=1;
			}
		}

		cursor = ptr;
		/* Uses cursor to advance along the blocks that have
		   been deallocated to calculate the total amount
		   of free adjacent memory. */
		tmp_flag=0;
		while  ( ( ( SHMEM_END - cursor ) > 0 ) && ( !tmp_flag ) )
		{
			if ( (HPTR cursor) < 0 )
			{
				cursor += ABS( (HPTR cursor) );
				cursor += HEADER_SIZE;
			}
			else
			{
				tmp_flag=1;
			}
		}

		/* Calcualte the total amount of free memory in this block */
		if ( (HPTR cursor) != LAST_SHMEM )
		{
			totsiz = (cursor - ptr) - HEADER_SIZE;
		}
		else
		{
			totsiz = ( (SHMEM_END - ptr) - HEADER_SIZE );
			/* if it's the last block we must exit from while */
			end = 1;
		}

		if ( totsiz >= memsiz ) //Found the right block
		{
			/* If th adjacent memory is greater than what we
			   need we split this memory in two blocks */
			if ( (totsiz - memsiz) > (int)HEADER_SIZE )
			{
				if ( (HPTR cursor) != LAST_SHMEM )
				{
					(HPTR (ptr + memsiz + HEADER_SIZE)) = - (totsiz - (memsiz + HEADER_SIZE) );
				}
				else
				{
					(HPTR (ptr + memsiz + HEADER_SIZE)) = LAST_SHMEM;
				}
				(HPTR ptr) = memsiz;
			}
			else // there is not enough space to split the block
			{
				(HPTR ptr) = totsiz;
			}
			end = 1;
			found = 1;
		}
	}

	if (found)
	{
		/* From 1 to shmem_start+1 */
		return (void *)( ptr + HEADER_SIZE - (shmem_start-1) );
	}
	else
	{
		return NULL;
	}
}

void sh_free(void *ptr) 
{
	char* p=(char *)ptr;
	HTYPE len;

	// Do nothing on a invalid pointer
	if ( !( (ptr > (void *)0) || (ptr < (void *)(shmem_size+2)) ) )
		return;

	p += (int)shmem_start + 1;

	p -= HEADER_SIZE; // Go back to read the header
	len = (HPTR p);
	if (len>0) /* Block used */
	{
		/* Go forward to read the header of the next block */
		char *tmp=p;
		tmp += ABS(len);
		tmp += HEADER_SIZE;


		/* If there aren't other blocks we can mark the start of
		   the just deallocated block as end of the allocated
		   heap. */
		if ( (SHMEM_END - tmp) > 0 )
		{
			if ( (HPTR tmp) == LAST_SHMEM )
				(HPTR p) = LAST_SHMEM;
			else
				HPTR p = -len;
		}
		else
		{
			HPTR p = LAST_SHMEM;
		}
		
	}
}


