#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sem.h>
#include <errno.h>
#include <time.h>
#include <logging.h>
#include <stdlib.h>

#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
	/* union semun is defined by including <sys/sem.h> */
#else
	/* according to X/OPEN we have to define it ourselves */
	union semun {
		int val;                    /* value for SETVAL */
		struct semid_ds *buf;       /* buffer for IPC_STAT, IPC_SET */
		unsigned short int *array;  /* array for GETALL, SETALL */
		struct seminfo *__buf;      /* buffer for IPC_INFO */
	};
#endif

int create_sem(int init_val) {
	int sem_id, i;
	union semun value;

	sem_id=-1;
	for (i=0;(i<20) && (sem_id==-1);i++) {
		unsigned int key;
		key=(unsigned int)(random());	
		sem_id=semget(key, 1, IPC_CREAT|IPC_EXCL|0700);
	}
	if (sem_id == -1) {
		log(LOG_ERR, "create_sem: cannot create semaphore set");
		//perror("semget");
		return -1;
	}

	value.val=init_val;
	if (semctl(sem_id, 0, SETVAL, value)==-1) {
		log(LOG_ERR, "create_sem: cannot set the init value");
		//perror("semctl");
		return -1;
	}
	return sem_id;
}

int remove_sem(int sem_id) {
	if (semctl(sem_id, 0, IPC_RMID, 0)==-1) {
		log(LOG_ERR, "remove_sem: cannot remove the semaphore set");
		//perror("remove_sem");
		return -1;
	}
	return 1;
}

int sem_P(int sem_id) {
	struct sembuf oper;
	int cycle = 1;

	oper.sem_num=0;
	oper.sem_op=-1;
	oper.sem_flg=0;

	while (cycle) {
		if (semop(sem_id, &oper, (unsigned int)1)==-1) {
			if (errno != EINTR) {
				log(LOG_ERR, "sem_P: cannot execute the P operation");
				//perror("semop");
				return -1;
			}
		}
		else
			cycle = 0;
	}
	return 1;
}

int sem_V(int sem_id) {
	struct sembuf oper;
	int cycle = 1;

	oper.sem_num=0;
	oper.sem_op=1;
	oper.sem_flg=0;

	while (cycle) {
		if (semop(sem_id, &oper, (unsigned int)1)==-1) {
			if (errno != EINTR) {
				log(LOG_ERR, "sem_V: cannot execute the V operation");
				//perror("semop");
				return -1;
			}
		}
		else
			cycle = 0;
	}
	return 1;
}
