#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include <netdb.h>

#include <errno.h>


#include <logging.h>

// Open *fd as a listening socket on the specified port
int tcp_listen(int *fd, unsigned short port)
{
	struct sockaddr_in sa;

	if ( (*fd=socket(AF_INET,SOCK_STREAM,0)) == -1 )
	{
		return -errno;
	}

	sa.sin_family=AF_INET;
	sa.sin_addr.s_addr=htonl(INADDR_ANY);
	sa.sin_port=htons(port);
	memset(&sa.sin_zero,0,sizeof(sa.sin_zero));

	if (bind(*fd,(struct sockaddr *)&sa,sizeof(struct sockaddr_in)) < 0)
	{
		return -errno;
	}

	if ( (listen(*fd,0)) == -1 )
	{
		return -errno;
	}

	return 0;
}

// Open *fd as a connection to the specified server and port. It uses
// time out (conn_timeout seconds).
int tcp_open(int *fd, char *server, unsigned short port, int conn_timeout)
{
	struct sockaddr_in sa;
	struct hostent *he;
	int flags;

	if ( (he=gethostbyname(server)) == NULL )
	{
		return -errno;
	}

	if ( ((*fd) = socket(AF_INET,SOCK_STREAM,0)) < 0 )
	{
		return -errno;
	}

	sa.sin_family=AF_INET;
	sa.sin_port=htons(port);
	memcpy(&sa.sin_addr,he->h_addr,he->h_length);
	memset(&sa.sin_zero,0,sizeof(sa.sin_zero));

	if ( (flags=fcntl(*fd,F_GETFL,0)) < 0 )
	{
		close(*fd);
		return -errno;
	}
	flags |= O_NONBLOCK;
	if ( (fcntl(*fd,F_SETFL,flags)) < 0 )
	{
		close(*fd);
		return -errno;
	}

	if (connect(*fd,(struct sockaddr *)&sa,sizeof(struct sockaddr_in)) < 0)
	{
		fd_set rset,wset;
		struct timeval timeout;

		if (errno != EINPROGRESS)
		{
			close(*fd);
			return -errno;
		}

		FD_ZERO(&rset);
		FD_ZERO(&wset);
		FD_SET(*fd,&rset);
		FD_SET(*fd,&wset);

		timeout.tv_sec = conn_timeout;
		timeout.tv_usec = 0;

		if ( select(*fd+1,&rset,&wset,NULL,&timeout) < 0 )
		{
			close(*fd);
			return -errno;
		}

		if ( (!FD_ISSET(*fd,&rset)) && (!FD_ISSET(*fd,&wset)) )
		{
			// timed out
			close(*fd);
			return -errno;
		}

		if (connect(*fd,(struct sockaddr *)&sa,sizeof(struct sockaddr_in)) < 0)
		{
			if (errno != EISCONN)
			{
				// connection failed
				close(*fd);
				return -errno;
			}
		}
	}

#ifndef NONBLOCKING_RW
	flags &= ~(O_NONBLOCK);
	if ( (fcntl(*fd,F_SETFL,flags)) < 0 )
	{
		close(*fd);
		return -errno;
	}
#endif

	return 0;
}

// Returns a string containing the IP of the host
char *resolve_host(char *host)
{
	// 20 chars are ok for an ipv4 address
	char ip[20];
	char *ret;
	struct hostent *he;

	if ( (he=gethostbyname(host)) == NULL )
	{
		log(LOG_WARNING,"resolve_host: cannot resolve %s, error %d",host,errno);
		return NULL;
	}

	snprintf(ip,sizeof(ip),"%d.%d.%d.%d",(unsigned char)he->h_addr[0],
	                                     (unsigned char)he->h_addr[1],
	                                     (unsigned char)he->h_addr[2],
	                                     (unsigned char)he->h_addr[3]);
	ip[sizeof(ip)-1]='\0';

	if ( (ret=malloc(strlen(ip)+1)) == NULL )
	{
		log(LOG_WARNING,"resolve_host: not enough memory");
		return NULL;
	}
	strcpy(ret,ip);
	return ret;
}

