#include <expire.h>
#include <hdr_parser.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <conv_date.h>
#include <buffers.h>

/* return the integer value of the string pointed by from */
int get_intvalue(char* from) {
	int i, ok, ret;

	i=ok=0;
	while ((from[i]!='\0')&&(ok==0)) {
		if (!isdigit(from[i]))
			i++;
		else ok=1;
	}
	ret=atoi(&from[i]);

	return ret;
}

/* return 1 if the doc is older than max-age value, 0 if not, -1 if the
   "s-maxage" and "max-age" directives are not present */
int too_old(char* header) {
	char *buf, *maxage_ptr, *smaxage_ptr;
	int doc_age, maxage_v, smaxage_v;	

	buf=parse_http_hdr("Cache-Control", header);
	if (buf==NULL)
		return -1;
	else {
		if ((smaxage_ptr=strstrcase(buf, "s-maxage"))==NULL) {
		/* if it does not contain the "s-maxage" directive */
				
			if ((maxage_ptr=strstrcase(buf, "max-age"))==NULL) {
			/* if it does not contain the "max-age" directive */
				free(buf);
            	return -1;
			}
			else {
				maxage_v=get_intvalue(maxage_ptr);
				free(buf);
				doc_age=compute_age(header);
				if ((doc_age!=-1)&&(maxage_v>=doc_age))
					return 0;
				else return 1;
			}
		}
		else {
			smaxage_v=get_intvalue(smaxage_ptr);
			free(buf);
			doc_age=compute_age(header);
			if ((doc_age!=-1)&&(smaxage_v>=doc_age))
				return 0;
			else return 1;
		}
	}
}

/* return 1 if the "Cache-Control" header contains the directive
   "must-revalidate" o "proxy-revalidate", 0 otherwise */
int to_revalidate(char* header) {
	char* buf;

	buf=parse_http_hdr("Cache-Control", header);
	if (buf==NULL) //does not contain the "Cache-Control" header
		return 0;
	else if (strstrcase(buf, "must-revalidate")==NULL) { /* not contain 
													   	    must-revalidate */
			if (strstrcase(buf, "proxy-revalidate")==NULL) { /* not contain 
														        proxy-revalidate
															 */
				free(buf);
				return 0;
			}
		}
	free(buf);
	return 1;
}

/* return 0 if the expire_value is greater than the actual date_value (in other
   words, if the document is valid), 1 otherwise */
int expire(char* header) {
	char* buf;
	struct tm doc_date;
	int doc_sec, our_sec;

	buf=parse_http_hdr("Expires", header);
	if (buf!=NULL) {
		doc_date=parse_date(buf);
		if (VALID(doc_date)) {
			doc_sec=(int)mktime(&doc_date);
			our_sec=(int)time(NULL);

			if (doc_sec>=our_sec) {
				free(buf);
				return 0;
			}
		}
		free(buf);
	}
	return 1;
}

int is_expired(char* header) {
	int ret;

	if (header==NULL)
		return -1;
	if (to_revalidate(header)==0) {
		if ((ret=too_old(header))==0) {
			return 0;
		}
		else if (ret==-1) {
				if (expire(header)==0) {
					return 0;
				}
			}
	}
	return 1;
}

int compute_age(char* header) {
	char* buf;
	int doc_sec, our_sec, age;
	struct tm doc_str;

	buf=parse_http_hdr("Date", header);
	if (buf!=NULL) {
		doc_str=parse_date(buf);
		if (VALID(doc_str)) {
			doc_sec=(int)mktime(&doc_str);
			our_sec=(int)time(NULL);
			age=our_sec-doc_sec;
			free(buf);
			if (age<0) age=0;
			return age;
		}
	}
	free(buf);
	return -1;
}
