# include <string.h>
# include <stdlib.h>
# include <time.h>
# include <hdr_parser.h>
# include <hdr_modify.h>
# include <buffers.h>
# include <const.h>
# include <conv_date.h>
# include <logging.h>

int prepare_validation_header (char *req_header,char **resp_header)
{
	char *etag_val=NULL;
	char *lastmod_val=NULL;
	char *tmp_val=NULL;
	int weak=0;
	int len_hdr = 0;
	char *append_hdr;
	char *tmp_check;

	if ( (tmp_check = parse_http_hdr("If-None-Match",*resp_header)) == NULL )
	{
		if ( (etag_val=parse_http_hdr("ETag",req_header)) != NULL )
		{
			if (strncmp (etag_val,"W/",2) )
			{
				weak = 1; /* weak validator */
			}
			len_hdr = strlen(etag_val) + 16;

			if ((append_hdr = malloc (len_hdr)) == NULL)
			{
				log(LOG_ERR,"prepare_validation_req:not enough memory");
				free(etag_val);
				return (-1);
			}

			strcpy (append_hdr,"If-None-Match: ");
			strcat (append_hdr,etag_val);
			free(etag_val);

			add_header (resp_header,append_hdr,1);
			free (append_hdr);
		}
	}
	else
	{
		free(tmp_check);
	}
	append_hdr = NULL;

	if ((tmp_check = parse_http_hdr("If-Modified-Since",*resp_header)) == NULL)
	{
		if ( (lastmod_val=parse_http_hdr("Last-Modified",req_header)) != NULL )
		{
			if ( (tmp_val=parse_http_hdr("Date",req_header)) != NULL )
			{
				struct tm start;
				struct tm end;

				start = parse_date (tmp_val);
				end = parse_date (lastmod_val);
				if ( abs (mktime(&start) - mktime (&end)) < 60 )
					weak = 1;
			}

			len_hdr = strlen (lastmod_val) + 20;
			if ( (append_hdr = (char *) malloc (len_hdr)) == NULL )
			{
				log(LOG_ERR,"prepare_validation_req:not enough memory");
				free(lastmod_val);
				free(append_hdr);
				free(tmp_val);
				return (-1);
			}

			strcpy (append_hdr,"If-Modified-Since: ");
			strcat (append_hdr,lastmod_val);
			free(lastmod_val);

			add_header (resp_header,append_hdr,1);
			free (append_hdr);
			free (tmp_val);
		}
	}
	else
	{
		free(tmp_check);
	}

	return (weak);
}

// If field cache-control is present in the header, call this one
// This function is only a template. Directive's behaviour is more complicated and different

int cache_response_control (char *directive)
{
	if ( strstrcase (directive,"private") != NULL )
		return (NO_CACHE);
	if ( strstrcase (directive,"no-cache") != NULL )
		return (NO_CACHE);
	if ( strstrcase (directive,"no-store") != NULL )
		return (NO_CACHE);
	return (CACHE);
}

